from typing import List

from myst.models import base_model
from myst.openapi.models.backtest_job_get import BacktestJobGet


class ProjectBacktestJobList(base_model.BaseModel):
    """Backtest job list schema."""

    data: List[BacktestJobGet]
    has_more: bool



from typing import TYPE_CHECKING, List, Optional, Union
from uuid import UUID

from myst import get_client
from myst.adapters.utils import get_resource_uuid
from myst.core.time.time import Time
from myst.models.types import UUIDOrStr
from myst.openapi.api.projects.deployments import (
    create_project_deployment,
    deactivate_project_deployment,
    list_project_deployments,
)
from myst.openapi.models.deployment_create import DeploymentCreate
from myst.resources.resource import Resource

if TYPE_CHECKING:  # Avoid circular imports.
    from myst.resources.project import Project


class Deployment(Resource):
    """A particular deployment of a project.

    A deployment is "active" if its `activate_time` is not `None` and its `deactivate_time` is `None`.

    Attributes:
        title: the title of the deployment.
        project: the identifier of the project with which the deployment is associated.
        activate_time: the time at which the deployment was activated, if any.
        deactive_time: the time at which the deployment was deactivated, if any.
    """

    title: str
    project: UUID
    activate_time: Optional[Time] = None
    deactivate_time: Optional[Time] = None

    @classmethod
    def create(cls, project: Union["Project", UUIDOrStr], title: str) -> "Deployment":
        """Creates a deployment for the project.

        Args:
            project: the project in which to create the deployment
            title: the title for the deployment

        Returns:
            the newly created deployment
        """
        deployment = create_project_deployment.request_sync(
            client=get_client(),
            project_uuid=str(get_resource_uuid(project)),
            json_body=DeploymentCreate(object="Deployment", title=title),
        )

        return Deployment.parse_obj(dict(deployment.dict(), project=get_resource_uuid(project)))

    @classmethod
    def get(cls, project: Union["Project", UUIDOrStr], uuid: UUIDOrStr) -> "Deployment":
        """Gets a specific deployment by its identifier."""
        deployments = cls.list(project)
        deployment = next((deployment for deployment in deployments if str(deployment.uuid) == str(uuid)), None)

        if deployment is None:
            raise ValueError(f"The deployment with {str(uuid)} can not found.")

        return deployment

    @classmethod
    def list(cls, project: Union["Project", UUIDOrStr]) -> List["Deployment"]:
        """Lists all deployments for specified project."""
        deployments = list_project_deployments.request_sync(
            client=get_client(), project_uuid=str(get_resource_uuid(project))
        )

        return [
            Deployment.parse_obj(dict(deployment.dict(), project=get_resource_uuid(project)))
            for deployment in deployments.data
        ]

    def deactivate(self) -> None:
        """Deactivate the deployment."""
        deployment = deactivate_project_deployment.request_sync(
            client=get_client(), project_uuid=str(self.project), uuid=str(self.uuid)
        )

        self.deactivate_time = Deployment.parse_obj(dict(deployment.dict(), project=self.project)).deactivate_time

    @property
    def is_active(self) -> bool:
        """Returns whether or not this deployment is active."""
        return self.deactivate_time is None



from myst.client import Client
from myst.openapi.models.time_series_run_result_list import TimeSeriesRunResultList


def request_sync(client: Client, project_uuid: str, time_series_uuid: str) -> TimeSeriesRunResultList:
    """Lists time series run results for the given time series."""

    return client.request(
        method="get",
        path=f"/projects/{project_uuid}/time_series/{time_series_uuid}/run_results/",
        response_class=TimeSeriesRunResultList,
    )



from typing import List, Union

from myst.models import base_model
from myst.openapi.models.model_fit_policy_get import ModelFitPolicyGet
from myst.openapi.models.time_series_run_policy_get import TimeSeriesRunPolicyGet


class ProjectNodePolicyList(base_model.BaseModel):
    """Schema for a list of project policies."""

    data: List[Union[TimeSeriesRunPolicyGet, ModelFitPolicyGet]]
    has_more: bool



from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.organization_role import OrganizationRole


class UserCreate(base_model.BaseModel):
    """Schema for user create requests."""

    object_: Literal["User"] = Field(..., alias="object")
    email: str
    organization: str
    organization_role: OrganizationRole



from typing import Any, Optional

import numpy as np

from myst.core.time.time import Time
from myst.core.time.time_delta import TimeDelta


def _num_periods_in_range(start_time: Time, end_time: Time, sample_period: TimeDelta) -> int:
    """Returns the number of periods in the given time range."""
    tdelta = end_time.to_pandas_timestamp() - start_time.to_pandas_timestamp()
    num_periods = tdelta / sample_period.to_pandas_frequency()
    if not num_periods.is_integer():
        raise ValueError("Period does not evenly divide time range.")

    return int(num_periods)


def check_arrays_same_length_as_time_range(
    start_time: Time, end_time: Time, sample_period: TimeDelta, values: np.ndarray, mask: Optional[np.ndarray]
) -> None:
    """Checks that values and mask arrays have the same length as the given time range."""
    num_periods = _num_periods_in_range(start_time=start_time, end_time=end_time, sample_period=sample_period)

    if len(values) != num_periods:
        raise ValueError("Values must have same length as time interval.")

    if mask is not None and len(mask) != num_periods:
        raise ValueError("Mask must have same length as time interval.")


def ensure_not_none(value: Any) -> Any:
    """Raises a ValueError if the given value is None."""
    if value is None:
        raise ValueError("Value cannot be `None`.")

    return value


def to_numpy_float_array(value: Any) -> np.ndarray:
    """Coerces the argument into a float-valued `np.ndarray` and returns it."""
    if not isinstance(value, np.ndarray) or value.dtype != float:
        return np.array(value, dtype=float)

    return value


def maybe_to_boolean_int_array(value: Any) -> Optional[np.ndarray]:
    """Checks that passed value contains only booleans or integers representing booleans and returns it."""
    if value is not None:
        if not isinstance(value, np.ndarray):
            value = np.array(value)

        if value.dtype not in (np.int8, int, bool) or not np.array_equal(value, value.astype(bool)):
            raise ValueError("Value must contain only boolean or boolean integer values.")

        return value.astype(np.int8)

    return value



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from myst.adapters.timing import to_timing_create
from myst.adapters.utils import drop_unset_values, get_resource_uuid
from myst.client import get_client
from myst.models.timing import AbsoluteOrRelativeTiming, TimeRangeBoundary, from_time_range_boundary
from myst.models.types import UNSET, ItemOrSlice, OptionalArgument, Unset, UUIDOrStr
from myst.openapi.api.projects.time_series.layers import (
    create_time_series_layer,
    get_time_series_layer,
    list_time_series_layers,
    update_time_series_layer,
)
from myst.openapi.models.layer_create import LayerCreate
from myst.openapi.models.layer_update import LayerUpdate
from myst.resources.edge import Edge

if TYPE_CHECKING:  # Avoid circular imports.
    from myst.resources.node import Node
    from myst.resources.project import Project
    from myst.resources.time_series import TimeSeries


class TimeSeriesLayer(Edge):
    """An edge into a time series.

    Time series layers are a way of stitching together data from multiple upstream nodes into a single, cohesive time
    series. Data can be combined across different time ranges, and a time series can use data from a lower-precedence
    layer when data is missing from a higher-precedence layer.

    Attributes:
        order: integer specifying priority of this layer when combining multiple layers; lower order implies higher
            precedence
        start_timing: the beginning of the natural time range this layer should produce data for; if None, there is no
            restriction on the beginning of the range
        end_timing: the end of the natural time range this layer should produce data for; if None, there is no
            restriction on the end of the range
    """

    order: int
    start_timing: Optional[AbsoluteOrRelativeTiming] = None
    end_timing: Optional[AbsoluteOrRelativeTiming] = None

    @classmethod
    def create(
        cls,
        project: Union["Project", UUIDOrStr],
        time_series: Union["TimeSeries", UUIDOrStr],
        node: Union["Node", UUIDOrStr],
        order: int,
        output_index: int = 0,
        label_indexer: Optional[ItemOrSlice] = None,
        start_timing: Optional[TimeRangeBoundary] = None,
        end_timing: Optional[TimeRangeBoundary] = None,
    ) -> "TimeSeriesLayer":
        """Creates a layer into this time series."""
        layer = create_time_series_layer.request_sync(
            client=get_client(),
            project_uuid=str(get_resource_uuid(project)),
            time_series_uuid=str(get_resource_uuid(time_series)),
            json_body=LayerCreate(
                object="Edge",
                type="Layer",
                upstream_node=str(get_resource_uuid(node)),
                order=order,
                output_index=output_index,
                label_indexer=label_indexer,
                start_timing=to_timing_create(from_time_range_boundary(start_timing)),
                end_timing=to_timing_create(from_time_range_boundary(end_timing)),
            ),
        )

        return TimeSeriesLayer.parse_obj(dict(layer.dict(), project=get_resource_uuid(project)))

    @classmethod
    def get(
        cls, project: Union["Project", UUIDOrStr], time_series: Union["TimeSeries", UUIDOrStr], uuid: UUIDOrStr
    ) -> "TimeSeriesLayer":
        """Gets a specific layer by its identifier."""
        layer = get_time_series_layer.request_sync(
            client=get_client(),
            project_uuid=str(get_resource_uuid(project)),
            time_series_uuid=str(get_resource_uuid(time_series)),
            uuid=str(uuid),
        )

        return TimeSeriesLayer.parse_obj(dict(layer.dict(), project=get_resource_uuid(project)))

    @classmethod
    def list(
        cls, project: Union["Project", UUIDOrStr], time_series: Union["TimeSeries", UUIDOrStr]
    ) -> List["TimeSeriesLayer"]:
        """Lists all layers for specified time series."""
        layers = list_time_series_layers.request_sync(
            client=get_client(),
            project_uuid=str(get_resource_uuid(project)),
            time_series_uuid=str(get_resource_uuid(time_series)),
        )

        # Parse the layers into their final representation.
        unsorted_layers = [
            TimeSeriesLayer.parse_obj(dict(layer.dict(), project=get_resource_uuid(project))) for layer in layers.data
        ]

        # The API returns the layers ordered by create time, descending. We prefer explicit layer ordering here.
        return sorted(unsorted_layers, key=lambda layer: layer.order)

    def update(
        self,
        node: OptionalArgument[Union["Node", UUIDOrStr]] = UNSET,
        order: OptionalArgument[int] = UNSET,
        output_index: OptionalArgument[int] = UNSET,
        label_indexer: OptionalArgument[ItemOrSlice] = UNSET,
        start_timing: OptionalArgument[TimeRangeBoundary] = UNSET,
        end_timing: OptionalArgument[TimeRangeBoundary] = UNSET,
    ) -> "TimeSeriesLayer":
        """Updates layer."""
        layer_update: Dict[str, Any] = dict(
            object="Edge",
            type="Layer",
            upstream_node=node if node is None or isinstance(node, Unset) else str(get_resource_uuid(node)),
            order=order,
            output_index=output_index,
            label_indexer=label_indexer,
            start_timing=to_timing_create(from_time_range_boundary(start_timing)),
            end_timing=to_timing_create(from_time_range_boundary(end_timing)),
        )

        layer = update_time_series_layer.request_sync(
            client=get_client(),
            project_uuid=str(self.project),
            time_series_uuid=str(self.downstream_node),
            uuid=str(self.uuid),
            json_body=LayerUpdate.parse_obj(drop_unset_values(layer_update)),
        )

        return TimeSeriesLayer.parse_obj(dict(layer.dict(), project=self.project))



from typing import List, Union

from myst.models import base_model
from myst.openapi.models.input_get import InputGet
from myst.openapi.models.layer_get import LayerGet


class ProjectEdgesList(base_model.BaseModel):
    """Project edges list schema."""

    data: List[Union[InputGet, LayerGet]]
    has_more: bool



from functools import singledispatch
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from myst.adapters.utils import drop_unset_values
from myst.client import get_client
from myst.connectors.model_connector import ModelConnector
from myst.connectors.operation_connector import OperationConnector
from myst.connectors.source_connector import SourceConnector
from myst.core.time.time_delta import TimeDelta
from myst.models.enums import DeployStatus, OrgSharingRole
from myst.models.types import UNSET, AxisLabels, CoordinateLabels, OptionalArgument, Shape, UUIDOrStr, to_uuid
from myst.openapi.api.projects import create_project, delete_project, get_project, list_projects, update_project
from myst.openapi.api.projects.edges import list_project_edges
from myst.openapi.api.projects.nodes import list_project_nodes
from myst.openapi.models.input_get import InputGet
from myst.openapi.models.layer_get import LayerGet
from myst.openapi.models.model_get import ModelGet
from myst.openapi.models.operation_get import OperationGet
from myst.openapi.models.org_sharing_role import OrgSharingRole as OAIOrgSharingRole
from myst.openapi.models.project_create import ProjectCreate
from myst.openapi.models.project_update import ProjectUpdate
from myst.openapi.models.source_get import SourceGet
from myst.openapi.models.time_series_get import TimeSeriesGet
from myst.resources.deployment import Deployment
from myst.resources.edge import Edge
from myst.resources.input import ModelInput, OperationInput
from myst.resources.layer import TimeSeriesLayer
from myst.resources.model import Model
from myst.resources.node import Node
from myst.resources.operation import Operation
from myst.resources.resource import ShareableResource
from myst.resources.source import Source
from myst.resources.time_series import TimeSeries

if TYPE_CHECKING:  # Avoid circular imports.
    from myst.recipes.time_series_recipe import TimeSeriesRecipe


@singledispatch
def _parse_node(node: Union[TimeSeriesGet, SourceGet, OperationGet, ModelGet]) -> Node:
    raise NotImplementedError()


@_parse_node.register
def _parse_time_series(node: TimeSeriesGet) -> TimeSeries:
    return TimeSeries.parse_obj(node.dict())


@_parse_node.register
def _parse_source(node: SourceGet) -> Source:
    return Source.parse_obj(node.dict())


@_parse_node.register
def _parse_model(node: ModelGet) -> Model:
    return Model.parse_obj(node.dict())


@_parse_node.register
def _parse_operation(node: OperationGet) -> Operation:
    return Operation.parse_obj(node.dict())


@singledispatch
def _parse_edge(edge: Union[InputGet, LayerGet], nodes: List[Node]) -> Edge:
    raise NotImplementedError()


@_parse_edge.register
def _parse_input(edge: InputGet, nodes: List[Node]) -> Union[ModelInput, OperationInput]:
    node = next((node for node in nodes if str(node.uuid) == edge.downstream_node), None)
    if node is None:
        raise ValueError("Could not find `downstream_node`.")
    if isinstance(node, Model):
        return ModelInput.parse_obj(dict(edge.dict(), project=node.project))
    else:
        return OperationInput.parse_obj(dict(edge.dict(), project=node.project))


@_parse_edge.register
def _parse_layer(edge: LayerGet, nodes: List[Node]) -> TimeSeriesLayer:
    node = next((node for node in nodes if str(node.uuid) == edge.downstream_node), None)
    if node is None:
        raise ValueError("Could not find `downstream_node`.")
    return TimeSeriesLayer.parse_obj(dict(edge.dict(), project=node.project))


class Project(ShareableResource):
    """A workspace for creating and deploying graphs of time series.

    A project collects the nodes and edges of the graph along with any policies used to execute the graph.
    """

    title: str
    deploy_status: DeployStatus
    description: Optional[str] = None

    @classmethod
    def create(
        cls,
        title: str,
        description: Optional[str] = None,
        organization_sharing_enabled: bool = False,
        organization_sharing_role: Optional[OrgSharingRole] = None,
    ) -> "Project":
        """Creates a new project.

        Args:
            title: the title of the project
            description: a brief description of the project
            organization_sharing_enabled: whether the project should be shared with the organization
            organization_sharing_role: the level of sharing access to this project to give users in the organization

        Returns:
            the newly created project
        """
        project = create_project.request_sync(
            client=get_client(),
            json_body=ProjectCreate(
                object="Project",
                title=title,
                description=description,
                organization_sharing_enabled=organization_sharing_enabled,
                organization_sharing_role=organization_sharing_role and OAIOrgSharingRole(organization_sharing_role),
            ),
        )

        return Project.parse_obj(project.dict())

    @classmethod
    def list(cls) -> List["Project"]:
        """Lists all projects visible to the user."""
        projects = list_projects.request_sync(client=get_client())

        return [Project.parse_obj(project.dict()) for project in projects.data]

    @classmethod
    def get(cls, uuid: UUIDOrStr) -> "Project":
        """Gets a specific project by its identifier."""
        project = get_project.request_sync(client=get_client(), uuid=str(to_uuid(uuid)))

        return Project.parse_obj(project.dict())

    def update(
        self,
        title: OptionalArgument[str] = UNSET,
        description: OptionalArgument[str] = UNSET,
        organization_sharing_enabled: OptionalArgument[bool] = UNSET,
        organization_sharing_role: OptionalArgument[OrgSharingRole] = UNSET,
    ) -> "Project":
        """Updates the project."""
        project_update: Dict[str, Any] = dict(
            object="Project",
            title=title,
            description=description,
            organization_sharing_enabled=organization_sharing_enabled,
            organization_sharing_role=organization_sharing_role and OAIOrgSharingRole(organization_sharing_role)
            if isinstance(organization_sharing_role, OrgSharingRole)
            else organization_sharing_role,
        )

        project = update_project.request_sync(
            client=get_client(),
            uuid=str(self.uuid),
            json_body=ProjectUpdate.parse_obj(drop_unset_values(project_update)),
        )

        return Project.parse_obj(project)

    def delete(self, requires_confirmation: bool = True) -> "Project":
        """Deletes the project.

        Args:
            requires_confirmation: whether or not the user is required to confirm the project title

        Raises:
            ValueError: if confirmation input title is not matched

        Returns:
            project that has been deleted
        """
        if requires_confirmation:
            title = input("Please input the project title to continue:")

            if title != self.title:
                raise ValueError("Input title is not matching the project title.")

        project = delete_project.request_sync(client=get_client(), uuid=str(self.uuid))

        return Project.parse_obj(project)

    @classmethod
    def filter(cls, title: str) -> List["Project"]:
        """Filters all projects visible to the user by title.

        Args:
            title: the title of the project to filter

        Returns:
            list of projects that have the specified title
        """
        projects = [project for project in cls.list() if project.title == title]

        return projects

    def list_nodes(self) -> List[Node]:
        """Lists all nodes in this project."""
        project_nodes = list_project_nodes.request_sync(client=get_client(), project_uuid=str(self.uuid))

        return [_parse_node(node) for node in project_nodes.data]

    def list_edges(self) -> List[Edge]:
        """Lists all edges in this project."""
        project_edges = list_project_edges.request_sync(client=get_client(), project_uuid=str(self.uuid))

        return [_parse_edge(edge, nodes=self.list_nodes()) for edge in project_edges.data]

    def create_source(self, title: str, connector: SourceConnector, description: Optional[str] = None) -> Source:
        """Creates a new source node in this project.

        Args:
            title: the title of the source
            connector: specification of the connector this source should use
            description: a brief description of the source

        Returns:
            the newly created source
        """
        return Source.create(project=self, title=title, connector=connector, description=description)

    def create_model(self, title: str, connector: ModelConnector, description: Optional[str] = None) -> Model:
        """Creates a new model node in this project.

        Args:
            title: the title of the model
            connector: specification of the connector this model should use
            description: a brief description of the model

        Returns:
            the newly created model
        """
        return Model.create(project=self, title=title, connector=connector, description=description)

    def create_operation(
        self, title: str, connector: OperationConnector, description: Optional[str] = None
    ) -> Operation:
        """Creates a new operation node in this project.

        Args:
            title: the title of the operation
            connector: specification of the connector this operation should use
            description: a brief description of the operation

        Returns:
            the newly created operation
        """
        return Operation.create(project=self, title=title, connector=connector, description=description)

    def create_time_series(
        self,
        title: str,
        sample_period: TimeDelta,
        cell_shape: Shape = (),
        coordinate_labels: CoordinateLabels = (),
        axis_labels: AxisLabels = (),
        description: Optional[str] = None,
    ) -> TimeSeries:
        """Creates a new time series node in this project.

        Args:
            title: the title of the time series
            sample_period: the frequency at which the data is sampled
            cell_shape: the shape of the data in each cell of the time series
            coordinate_labels: the labels for each coordinate of a cell in the time series
            axis_labels: the labels for each axis of a cell in the time series
            description: a brief description of the time series

        Returns:
            the newly created time series
        """
        return TimeSeries.create(
            project=self,
            title=title,
            sample_period=sample_period,
            cell_shape=cell_shape,
            coordinate_labels=coordinate_labels,
            axis_labels=axis_labels,
            description=description,
        )

    def create_time_series_from_recipe(
        self, recipe: "TimeSeriesRecipe", title: Optional[str] = None, description: Optional[str] = None
    ) -> TimeSeries:
        """Creates a new time series node in this project based on a provided recipe.

        Args:
            recipe: the recipe used to create the time series
            title: the title of the time series; if none is given, the recipe will supply a sensible default
            description: the description of the time series

        Returns:
            the newly created time series
        """
        return recipe.create(project=self, title=title, description=description)

    def deploy(self, title: str) -> Deployment:
        """Deploys the current project.

        Args:
            title: the title for the deployment

        Returns:
            the newly created deployment
        """
        return Deployment.create(project=self, title=title)



from typing import Optional, Union

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.absolute_timing_create import AbsoluteTimingCreate
from myst.openapi.models.cron_timing_create import CronTimingCreate
from myst.openapi.models.relative_timing_create import RelativeTimingCreate


class BacktestCreate(base_model.BaseModel):
    """Schema for backtest create requests."""

    title: str
    test_start_time: str
    test_end_time: str
    fit_start_timing: Union[AbsoluteTimingCreate, RelativeTimingCreate]
    fit_end_timing: Union[AbsoluteTimingCreate, RelativeTimingCreate]
    fit_reference_timing: Union[AbsoluteTimingCreate, CronTimingCreate]
    predict_start_timing: RelativeTimingCreate
    predict_end_timing: RelativeTimingCreate
    predict_reference_timing: CronTimingCreate
    model: str
    object_: Optional[Literal["Backtest"]] = Field("Backtest", alias="object")
    description: Optional[str] = None



from myst.client import Client
from myst.openapi.models.model_fit_job_get import ModelFitJobGet
from myst.openapi.models.node_job_create import NodeJobCreate


def request_sync(client: Client, project_uuid: str, uuid: str, json_body: NodeJobCreate) -> ModelFitJobGet:
    """Create an ad hoc fit job for a model."""

    return client.request(
        method="post",
        path=f"/projects/{project_uuid}/models/{uuid}:fit",
        response_class=ModelFitJobGet,
        request_model=json_body,
    )



from typing import List, Optional, Union

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.time_dataset_metadata import TimeDatasetMetadata
from myst.openapi.models.time_dataset_row import TimeDatasetRow


class TimeDataset(base_model.BaseModel):
    """"""

    object_: Literal["TimeDataset"] = Field(..., alias="object")
    cell_shape: List[int]
    sample_period: str
    data: List[TimeDatasetRow]
    coordinate_labels: Optional[List[List[Union[int, str]]]] = None
    axis_labels: Optional[List[Union[int, str]]] = None
    metadata: Optional[TimeDatasetMetadata] = None



from myst.client import Client
from myst.openapi.models.organization_list import OrganizationList


def request_sync(client: Client) -> OrganizationList:
    """Lists organizations."""

    return client.request(method="get", path=f"/organizations/", response_class=OrganizationList)



from myst.client import Client
from myst.openapi.models.model_fit_result_get import ModelFitResultGet


def request_sync(client: Client, project_uuid: str, model_uuid: str, uuid: str) -> ModelFitResultGet:
    """Gets the model fit result associated with the given identifier."""

    return client.request(
        method="get",
        path=f"/projects/{project_uuid}/models/{model_uuid}/fit_results/{uuid}",
        response_class=ModelFitResultGet,
    )



from myst.client import Client
from myst.openapi.models.input_get import InputGet


def request_sync(client: Client, project_uuid: str, operation_uuid: str, uuid: str) -> InputGet:
    """Deletes an input for an operation."""

    return client.request(
        method="delete",
        path=f"/projects/{project_uuid}/operations/{operation_uuid}/inputs/{uuid}",
        response_class=InputGet,
    )



from typing import Any, Dict

from myst.models import base_model


class OperationUpdateInputSpecsSchema(base_model.BaseModel):
    """**Ignored**. A JSON Schema definition describing the inputs that this operation can be used with. Output only."""

    __root__: Dict[str, Any]

    def __getitem__(self, item: str) -> Any:
        return self.__root__[item]



from typing import List, Optional

from myst.models.base_model import BaseModel


class ValidationError(BaseModel):
    """Model for validation error details returned by the server."""

    loc: List[str]
    msg: str
    type: str


class HTTPValidationError(BaseModel):
    """Model for a validation error (HTTP 422) returned by the API."""

    detail: Optional[List[ValidationError]] = None


class StringDetailError(BaseModel):
    """Pydantic model used to describe a simple string detail error."""

    detail: str



from typing import Any, Dict, Optional, Union

from myst.client import Client
from myst.openapi.models.connector_type import ConnectorType
from myst.openapi.models.polymorphic_connector_list import PolymorphicConnectorList

from ...types import UNSET, Unset


def request_sync(client: Client, type: Optional[ConnectorType] = None) -> PolymorphicConnectorList:
    """Lists all connectors or a specified type of connectors."""

    params: Dict[str, Any] = {}
    json_type: Union[Unset, None, str] = UNSET
    if not isinstance(type, Unset):
        json_type = type.value if type else None

    params["type"] = json_type

    params = {k: v for k, v in params.items() if v is not None}

    return client.request(method="get", path=f"/connectors/", response_class=PolymorphicConnectorList, params=params)



import pathlib
from typing import Optional

import pydantic

PROJECT_ROOT = pathlib.Path(__file__).parent.parent.absolute()


class Settings(pydantic.BaseSettings):
    """Runtime settings for the Myst client."""

    MYST_API_HOST: str = "https://api.myst.ai"

    # Increase default timeout, since it's insufficient for some longer inserts/retrievals.
    API_TIMEOUT_SEC: int = 30

    MYST_APPLICATION_CREDENTIALS: Optional[pathlib.Path] = None

    @pydantic.validator("MYST_APPLICATION_CREDENTIALS")
    def validate_service_account_key_path(cls, v: Optional[str]) -> Optional[pathlib.Path]:
        """Makes sure the service account key is a valid path."""
        if isinstance(v, pathlib.Path):
            return v
        elif isinstance(v, str):
            return pathlib.Path(v)
        elif v is None:
            return None
        else:
            raise ValueError("`MYST_APPLICATION_CREDENTIALS` must be a string if set.")


settings = Settings()



from typing import List, Optional

from myst.models import base_model
from myst.openapi.models.validation_error import ValidationError


class HTTPValidationError(base_model.BaseModel):
    """"""

    detail: Optional[List[ValidationError]] = None



from myst.client import Client
from myst.openapi.models.backtest_job_get import BacktestJobGet


def request_sync(client: Client, project_uuid: str, backtest_uuid: str, uuid: str) -> BacktestJobGet:
    """List all backtest jobs for the project."""

    return client.request(
        method="get",
        path=f"/projects/{project_uuid}/backtests/{backtest_uuid}/jobs/{uuid}",
        response_class=BacktestJobGet,
    )



from typing import TYPE_CHECKING, Any, Dict, List, Union
from uuid import UUID

from myst import Time, TimeDelta
from myst.adapters.timing import to_timing_create
from myst.adapters.utils import UNSET, drop_unset_values, get_resource_uuid
from myst.client import get_client
from myst.models.timing import (
    AbsoluteOrRelativeTiming,
    ScheduleTiming,
    TimeRangeBoundary,
    from_schedule_specifier,
    from_time_range_boundary,
)
from myst.models.types import OptionalArgument, UUIDOrStr
from myst.openapi.api.projects.models.fit_policies import (
    create_model_fit_policy,
    get_model_fit_policy,
    list_model_fit_policies,
    update_model_fit_policy,
)
from myst.openapi.api.projects.time_series.run_policies import (
    create_time_series_run_policy,
    get_time_series_run_policy,
    list_time_series_run_policies,
    update_time_series_run_policy,
)
from myst.openapi.models.model_fit_policy_create import ModelFitPolicyCreate
from myst.openapi.models.model_fit_policy_update import ModelFitPolicyUpdate
from myst.openapi.models.time_series_run_policy_create import TimeSeriesRunPolicyCreate
from myst.openapi.models.time_series_run_policy_update import TimeSeriesRunPolicyUpdate
from myst.resources.resource import Resource

if TYPE_CHECKING:  # Avoid circular imports.
    from myst.resources.model import Model
    from myst.resources.project import Project
    from myst.resources.time_series import TimeSeries


class Policy(Resource):
    """Describes when and over what natural time range to run a particular type of job for a node.

    Attributes:
        project: the identifier of the project this policy belongs to
        creator: the identifier of the user who created this resource
        schedule_timing: when the policy is scheduled to run, whether recurrent or once
        active: whether this policy is currently in effect
        node: the identifier of the node this policy applies to
        start_timing: the beginning of the natural time range for which this policy applies, inclusive
        end_timing: the end of the natural time range for which this policy applies, exclusive
    """

    project: UUID
    creator: UUID
    schedule_timing: ScheduleTiming
    active: bool
    node: UUID
    start_timing: AbsoluteOrRelativeTiming
    end_timing: AbsoluteOrRelativeTiming


class ModelFitPolicy(Policy):
    """Describes when and over what natural time range to run a fit job on a model."""

    @classmethod
    def create(
        cls,
        project: Union["Project", UUIDOrStr],
        model: Union["Model", UUIDOrStr],
        schedule_timing: Union[Time, TimeDelta, ScheduleTiming],
        start_timing: TimeRangeBoundary,
        end_timing: TimeRangeBoundary,
        active: bool = True,
    ) -> "ModelFitPolicy":
        """Creates a fit policy for the model."""
        model_fit_policy_create = ModelFitPolicyCreate(
            object="Policy",
            type="ModelFitPolicy",
            schedule_timing=to_timing_create(from_schedule_specifier(schedule_timing)),
            start_timing=to_timing_create(from_time_range_boundary(start_timing)),
            end_timing=to_timing_create(from_time_range_boundary(end_timing)),
            active=active,
        )

        model_fit_policy = create_model_fit_policy.request_sync(
            client=get_client(),
            project_uuid=str(get_resource_uuid(project)),
            model_uuid=str(get_resource_uuid(model)),
            json_body=model_fit_policy_create,
        )

        return ModelFitPolicy.parse_obj(dict(model_fit_policy.dict(), project=get_resource_uuid(project)))

    @classmethod
    def get(
        cls, project: Union["Project", UUIDOrStr], model: Union["Model", UUIDOrStr], uuid: UUIDOrStr
    ) -> "ModelFitPolicy":
        """Gets a specific fit policy by its identifier."""
        model_fit_policy = get_model_fit_policy.request_sync(
            client=get_client(),
            project_uuid=str(get_resource_uuid(project)),
            model_uuid=str(get_resource_uuid(model)),
            uuid=str(uuid),
        )

        return ModelFitPolicy.parse_obj(dict(model_fit_policy.dict(), project=get_resource_uuid(project)))

    @classmethod
    def list(cls, project: Union["Project", UUIDOrStr], model: Union["Model", UUIDOrStr]) -> List["ModelFitPolicy"]:
        """Lists all fit policies for the model."""
        model_fit_policies = list_model_fit_policies.request_sync(
            client=get_client(), project_uuid=str(get_resource_uuid(project)), model_uuid=str(get_resource_uuid(model))
        )

        return [
            ModelFitPolicy.parse_obj(dict(model_fit_policy.dict(), project=get_resource_uuid(project)))
            for model_fit_policy in model_fit_policies.data
        ]

    def update(
        self,
        schedule_timing: OptionalArgument[Union[Time, TimeDelta, ScheduleTiming]] = UNSET,
        start_timing: OptionalArgument[TimeRangeBoundary] = UNSET,
        end_timing: OptionalArgument[TimeRangeBoundary] = UNSET,
        active: OptionalArgument[bool] = UNSET,
    ) -> "ModelFitPolicy":
        """Updates the fit policy for the model."""
        model_fit_policy_update: Dict[str, Any] = dict(
            object="Policy",
            type="ModelFitPolicy",
            schedule_timing=to_timing_create(from_schedule_specifier(schedule_timing)),
            start_timing=to_timing_create(from_time_range_boundary(start_timing)),
            end_timing=to_timing_create(from_time_range_boundary(end_timing)),
            active=active,
        )

        model_fit_policy = update_model_fit_policy.request_sync(
            client=get_client(),
            project_uuid=str(self.project),
            model_uuid=str(self.node),
            uuid=str(self.uuid),
            json_body=ModelFitPolicyUpdate.parse_obj(drop_unset_values(model_fit_policy_update)),
        )

        return ModelFitPolicy.parse_obj(dict(model_fit_policy.dict(), project=self.project))


class TimeSeriesRunPolicy(Policy):
    """Describes when and over what natural time range to run a fit job on a model."""

    @classmethod
    def create(
        cls,
        project: Union["Project", UUIDOrStr],
        time_series: Union["TimeSeries", UUIDOrStr],
        schedule_timing: Union[Time, TimeDelta, ScheduleTiming],
        start_timing: TimeRangeBoundary,
        end_timing: TimeRangeBoundary,
        active: bool = True,
    ) -> "TimeSeriesRunPolicy":
        """Creates a run policy for the time series."""
        time_series_run_policy_create = TimeSeriesRunPolicyCreate(
            object="Policy",
            type="TimeSeriesRunPolicy",
            schedule_timing=to_timing_create(from_schedule_specifier(schedule_timing)),
            start_timing=to_timing_create(from_time_range_boundary(start_timing)),
            end_timing=to_timing_create(from_time_range_boundary(end_timing)),
            active=active,
        )

        time_series_run_policy = create_time_series_run_policy.request_sync(
            client=get_client(),
            project_uuid=str(get_resource_uuid(project)),
            time_series_uuid=str(get_resource_uuid(time_series)),
            json_body=time_series_run_policy_create,
        )

        return TimeSeriesRunPolicy.parse_obj(dict(time_series_run_policy.dict(), project=get_resource_uuid(project)))

    @classmethod
    def get(
        cls, project: Union["Project", UUIDOrStr], time_series: Union["TimeSeries", UUIDOrStr], uuid: UUIDOrStr
    ) -> "TimeSeriesRunPolicy":
        """Gets a specific run policy by its identifier."""
        time_series_run_policy = get_time_series_run_policy.request_sync(
            client=get_client(),
            project_uuid=str(get_resource_uuid(project)),
            time_series_uuid=str(get_resource_uuid(time_series)),
            uuid=str(uuid),
        )

        return TimeSeriesRunPolicy.parse_obj(dict(time_series_run_policy.dict(), project=get_resource_uuid(project)))

    @classmethod
    def list(
        cls, project: Union["Project", UUIDOrStr], time_series: Union["TimeSeries", UUIDOrStr]
    ) -> List["TimeSeriesRunPolicy"]:
        """Lists all run policies for the time series."""
        time_series_run_policies = list_time_series_run_policies.request_sync(
            client=get_client(),
            project_uuid=str(get_resource_uuid(project)),
            time_series_uuid=str(get_resource_uuid(time_series)),
        )

        return [
            TimeSeriesRunPolicy.parse_obj(dict(time_series_run_policy.dict(), project=get_resource_uuid(project)))
            for time_series_run_policy in time_series_run_policies.data
        ]

    def update(
        self,
        schedule_timing: OptionalArgument[Union[Time, TimeDelta, ScheduleTiming]] = UNSET,
        start_timing: OptionalArgument[TimeRangeBoundary] = UNSET,
        end_timing: OptionalArgument[TimeRangeBoundary] = UNSET,
        active: OptionalArgument[bool] = UNSET,
    ) -> "TimeSeriesRunPolicy":
        """Updates the run policy for the time series."""
        time_series_run_policy_update: Dict[str, Any] = dict(
            object="Policy",
            type="TimeSeriesRunPolicy",
            schedule_timing=to_timing_create(from_schedule_specifier(schedule_timing)),
            start_timing=to_timing_create(from_time_range_boundary(start_timing)),
            end_timing=to_timing_create(from_time_range_boundary(end_timing)),
            active=active,
        )

        time_series_run_policy = update_time_series_run_policy.request_sync(
            client=get_client(),
            project_uuid=str(self.project),
            time_series_uuid=str(self.node),
            uuid=str(self.uuid),
            json_body=TimeSeriesRunPolicyUpdate.parse_obj(drop_unset_values(time_series_run_policy_update)),
        )

        return TimeSeriesRunPolicy.parse_obj(dict(time_series_run_policy.dict(), project=self.project))



import enum
from typing import Optional
from uuid import UUID

from myst.connectors.model_connector import ModelConnector


@enum.unique
class GroupName(str, enum.Enum):
    FEATURES = "features"
    TARGETS = "targets"
    SAMPLE_WEIGHTS = "sample_weights"


@enum.unique
class XGBoostObjective(str, enum.Enum):
    REG_SQUARED_ERROR = "reg:squarederror"
    BINARY_LOGISTIC = "binary:logistic"


class XGBoost(ModelConnector):
    def __init__(
        self,
        objective: Optional[XGBoostObjective] = None,
        num_boost_round: Optional[int] = None,
        max_depth: Optional[int] = None,
        min_child_weight: Optional[int] = None,
        learning_rate: Optional[float] = None,
        subsample: Optional[float] = None,
        colsample_bytree: Optional[float] = None,
        colsample_bylevel: Optional[float] = None,
        colsample_bynode: Optional[float] = None,
        gamma: Optional[float] = None,
        reg_alpha: Optional[float] = None,
        reg_lambda: Optional[float] = None,
        fit_on_null_values: Optional[bool] = None,
        predict_on_null_values: Optional[bool] = None,
    ) -> None:
        super().__init__(
            uuid=UUID("b78ff94a-27b1-4986-a88a-536661239bb2"),
            parameters=dict(
                colsample_bylevel=colsample_bylevel,
                colsample_bynode=colsample_bynode,
                colsample_bytree=colsample_bytree,
                fit_on_null_values=fit_on_null_values,
                gamma=gamma,
                learning_rate=learning_rate,
                max_depth=max_depth,
                min_child_weight=min_child_weight,
                num_boost_round=num_boost_round,
                objective=objective,
                predict_on_null_values=predict_on_null_values,
                reg_alpha=reg_alpha,
                reg_lambda=reg_lambda,
                subsample=subsample,
            ),
        )



from myst.client import Client
from myst.openapi.models.project_backtest_job_list import ProjectBacktestJobList


def request_sync(client: Client, project_uuid: str, backtest_uuid: str) -> ProjectBacktestJobList:
    """List all backtest jobs for the backtest."""

    return client.request(
        method="get",
        path=f"/projects/{project_uuid}/backtests/{backtest_uuid}/jobs/",
        response_class=ProjectBacktestJobList,
    )



from myst.client import Client
from myst.openapi.models.organization_create import OrganizationCreate
from myst.openapi.models.organization_get import OrganizationGet


def request_sync(client: Client, json_body: OrganizationCreate) -> OrganizationGet:
    """Creates an organization."""

    return client.request(
        method="post", path=f"/organizations/", response_class=OrganizationGet, request_model=json_body
    )



from myst.client import Client
from myst.openapi.models.input_create import InputCreate
from myst.openapi.models.input_get import InputGet


def request_sync(client: Client, project_uuid: str, model_uuid: str, json_body: InputCreate) -> InputGet:
    """Creates an input for a model."""

    return client.request(
        method="post",
        path=f"/projects/{project_uuid}/models/{model_uuid}/inputs/",
        response_class=InputGet,
        request_model=json_body,
    )



from typing import List

from myst.models import base_model
from myst.openapi.models.organization_get import OrganizationGet


class OrganizationList(base_model.BaseModel):
    """Schema for organization list responses."""

    data: List[OrganizationGet]
    has_more: bool



from enum import Enum


class HPOMetric(str, Enum):
    MAE = "mae"
    MAPE = "mape"
    MEAN_NEGATIVE_LOG_LIKELIHOOD = "mean_negative_log_likelihood"
    MEAN_PINBALL_LOSS = "mean_pinball_loss"
    MSE = "mse"
    RMSE = "rmse"

    def __str__(self) -> str:
        return str(self.value)



from typing import Optional, Union

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.absolute_timing_get import AbsoluteTimingGet
from myst.openapi.models.cron_timing_get import CronTimingGet
from myst.openapi.models.relative_timing_get import RelativeTimingGet


class BacktestGet(base_model.BaseModel):
    """Schema for backtest get responses."""

    object_: Literal["Backtest"] = Field(..., alias="object")
    uuid: str
    create_time: str
    title: str
    test_start_time: str
    test_end_time: str
    fit_start_timing: Union[AbsoluteTimingGet, RelativeTimingGet]
    fit_end_timing: Union[AbsoluteTimingGet, RelativeTimingGet]
    fit_reference_timing: Union[AbsoluteTimingGet, CronTimingGet]
    predict_start_timing: RelativeTimingGet
    predict_end_timing: RelativeTimingGet
    predict_reference_timing: CronTimingGet
    project: str
    model: str
    creator: str
    update_time: Optional[str] = None
    description: Optional[str] = None



from typing import Any, Dict, Optional

from myst.client import Client
from myst.openapi.models.user_list import UserList


def request_sync(client: Client, organization_uuid: Optional[str] = None) -> UserList:
    """Lists users."""

    params: Dict[str, Any] = {}
    params["organization_uuid"] = organization_uuid

    params = {k: v for k, v in params.items() if v is not None}

    return client.request(method="get", path=f"/users/", response_class=UserList, params=params)



from myst.client import Client
from myst.openapi.models.backtest_get import BacktestGet
from myst.openapi.models.backtest_update import BacktestUpdate


def request_sync(client: Client, project_uuid: str, uuid: str, json_body: BacktestUpdate) -> BacktestGet:
    """Updates a backtest."""

    return client.request(
        method="patch",
        path=f"/projects/{project_uuid}/backtests/{uuid}",
        response_class=BacktestGet,
        request_model=json_body,
    )



from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model


class AbsoluteTimingGet(base_model.BaseModel):
    """Absolute timing schema for get responses."""

    object_: Literal["Timing"] = Field(..., alias="object")
    type: Literal["AbsoluteTiming"]
    time: str



from myst.client import Client
from myst.openapi.models.permission_list import PermissionList
from myst.openapi.models.shareable_resource_share import ShareableResourceShare


def request_sync(client: Client, project_uuid: str, uuid: str, json_body: ShareableResourceShare) -> PermissionList:
    """Shares a time series."""

    return client.request(
        method="post",
        path=f"/projects/{project_uuid}/time_series/{uuid}:share",
        response_class=PermissionList,
        request_model=json_body,
    )



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from myst.adapters.utils import drop_unset_values, get_resource_uuid
from myst.client import get_client
from myst.models.types import UNSET, ItemOrSlice, OptionalArgument, Unset, UUIDOrStr
from myst.openapi.api.projects.models.inputs import (
    create_model_input,
    get_model_input,
    list_model_inputs,
    update_model_input,
)
from myst.openapi.api.projects.operations.inputs import (
    create_operation_input,
    get_operation_input,
    list_operation_inputs,
    update_operation_input,
)
from myst.openapi.models.input_create import InputCreate
from myst.openapi.models.input_update import InputUpdate
from myst.resources.edge import Edge

if TYPE_CHECKING:  # Avoid circular imports.
    from myst.resources.model import Model
    from myst.resources.operation import Operation
    from myst.resources.project import Project
    from myst.resources.time_series import TimeSeries


class ModelInput(Edge):
    """An edge from a time series into a model."""

    @classmethod
    def create(
        cls,
        project: Union["Project", UUIDOrStr],
        model: Union["Model", UUIDOrStr],
        time_series: Union["TimeSeries", UUIDOrStr],
        group_name: str,
        output_index: int = 0,
        label_indexer: Optional[ItemOrSlice] = None,
    ) -> "ModelInput":
        """Creates a new model input edge between the given nodes.

        Args:
            project: the project this edge belongs to
            model: the model into which data flows out of this edge
            time_series: the time series from which data flows into this edge
            group_name: the name of the group of inputs on the underlying model or operation connector to which to pass
                the data from this input
            output_index: which time dataset, out of the sequence of upstream time datasets, to pass to the downstream
                node
            label_indexer: the slice of the upstream data to pass to the downstream node

        Returns:
            the newly created model input
        """
        input_create = InputCreate(
            object="Edge",
            type="Input",
            upstream_node=str(get_resource_uuid(time_series)),
            group_name=group_name,
            output_index=output_index,
            label_indexer=label_indexer,
        )

        input_ = create_model_input.request_sync(
            client=get_client(),
            project_uuid=str(get_resource_uuid(project)),
            model_uuid=str(get_resource_uuid(model)),
            json_body=input_create,
        )

        return ModelInput.parse_obj(dict(input_.dict(), project=get_resource_uuid(project)))

    @classmethod
    def get(
        cls, project: Union["Project", UUIDOrStr], model: Union["Model", UUIDOrStr], uuid: UUIDOrStr
    ) -> "ModelInput":
        """Gets a specific model input by its identifier."""
        input_ = get_model_input.request_sync(
            client=get_client(),
            project_uuid=str(get_resource_uuid(project)),
            model_uuid=str(get_resource_uuid(model)),
            uuid=str(uuid),
        )

        return ModelInput.parse_obj(dict(input_.dict(), project=get_resource_uuid(project)))

    @classmethod
    def list(cls, project: Union["Project", UUIDOrStr], model: Union["Model", UUIDOrStr]) -> List["ModelInput"]:
        """Lists all model inputs for specified model node."""
        inputs = list_model_inputs.request_sync(
            client=get_client(), project_uuid=str(get_resource_uuid(project)), model_uuid=str(get_resource_uuid(model))
        )

        return [ModelInput.parse_obj(dict(input_.dict(), project=get_resource_uuid(project))) for input_ in inputs.data]

    def update(
        self,
        time_series: OptionalArgument[Union["TimeSeries", UUIDOrStr]] = UNSET,
        group_name: OptionalArgument[str] = UNSET,
        output_index: OptionalArgument[int] = UNSET,
        label_indexer: OptionalArgument[ItemOrSlice] = UNSET,
    ) -> "ModelInput":
        """Updates the model input."""
        input_update: Dict[str, Any] = dict(
            object="Edge",
            type="Input",
            upstream_node=time_series
            if time_series is None or isinstance(time_series, Unset)
            else str(get_resource_uuid(time_series)),
            group_name=group_name,
            output_index=output_index,
            label_indexer=label_indexer,
        )

        input_ = update_model_input.request_sync(
            client=get_client(),
            project_uuid=str(self.project),
            model_uuid=str(self.downstream_node),
            uuid=str(self.uuid),
            json_body=InputUpdate.parse_obj(drop_unset_values(input_update)),
        )

        return ModelInput.parse_obj(dict(input_.dict(), project=self.project))


class OperationInput(Edge):
    """An edge from a time series into an operation."""

    @classmethod
    def create(
        cls,
        project: Union["Project", UUIDOrStr],
        operation: Union["Operation", UUIDOrStr],
        time_series: Union["TimeSeries", UUIDOrStr],
        group_name: str,
        output_index: int = 0,
        label_indexer: Optional[ItemOrSlice] = None,
    ) -> "OperationInput":
        """Creates a new operation input edge between the given nodes.

        Args:
            project: the project this edge belongs to
            operation: the operation into which data flows out of this edge
            time_series: the time series from which data flows into this edge
            group_name: the name of the group of inputs on the underlying model or operation connector to which to pass
                the data from this input
            output_index: which time dataset, out of the sequence of upstream time datasets, to pass to the downstream
                node
            label_indexer: the slice of the upstream data to pass to the downstream node

        Returns:
            the newly created operation input
        """
        input_create = InputCreate(
            object="Edge",
            type="Input",
            upstream_node=str(get_resource_uuid(time_series)),
            group_name=group_name,
            output_index=output_index,
            label_indexer=label_indexer,
        )

        input_ = create_operation_input.request_sync(
            client=get_client(),
            project_uuid=str(get_resource_uuid(project)),
            operation_uuid=str(get_resource_uuid(operation)),
            json_body=input_create,
        )

        return OperationInput.parse_obj(dict(input_.dict(), project=get_resource_uuid(project)))

    @classmethod
    def get(
        cls, project: Union["Project", UUIDOrStr], operation: Union["Operation", UUIDOrStr], uuid: UUIDOrStr
    ) -> "OperationInput":
        """Gets a specific operation input by its identifier."""
        input_ = get_operation_input.request_sync(
            client=get_client(),
            project_uuid=str(get_resource_uuid(project)),
            operation_uuid=str(get_resource_uuid(operation)),
            uuid=str(uuid),
        )

        return OperationInput.parse_obj(dict(input_.dict(), project=get_resource_uuid(project)))

    @classmethod
    def list(
        cls, project: Union["Project", UUIDOrStr], operation: Union["Operation", UUIDOrStr]
    ) -> List["OperationInput"]:
        """Lists all operation inputs for specified operation node."""
        inputs = list_operation_inputs.request_sync(
            client=get_client(),
            project_uuid=str(get_resource_uuid(project)),
            operation_uuid=str(get_resource_uuid(operation)),
        )

        return [
            OperationInput.parse_obj(dict(input_.dict(), project=get_resource_uuid(project))) for input_ in inputs.data
        ]

    def update(
        self,
        time_series: OptionalArgument[Union["TimeSeries", UUIDOrStr]] = UNSET,
        group_name: OptionalArgument[str] = UNSET,
        output_index: OptionalArgument[int] = UNSET,
        label_indexer: OptionalArgument[ItemOrSlice] = UNSET,
    ) -> "OperationInput":
        """Updates the operation input."""
        input_update: Dict[str, Any] = dict(
            object="Edge",
            type="Input",
            upstream_node=time_series
            if time_series is None or isinstance(time_series, Unset)
            else str(get_resource_uuid(time_series)),
            group_name=group_name,
            output_index=output_index,
            label_indexer=label_indexer,
        )

        input_ = update_operation_input.request_sync(
            client=get_client(),
            project_uuid=str(self.project),
            operation_uuid=str(self.downstream_node),
            uuid=str(self.uuid),
            json_body=InputUpdate.parse_obj(drop_unset_values(input_update)),
        )

        return OperationInput.parse_obj(dict(input_.dict(), project=self.project))



from typing import List

from myst.models import base_model
from myst.openapi.models.permission_get import PermissionGet


class PermissionList(base_model.BaseModel):
    """Schema for permission list responses."""

    data: List[PermissionGet]
    has_more: bool



from myst.client import Client
from myst.openapi.models.input_list import InputList


def request_sync(client: Client, project_uuid: str, model_uuid: str) -> InputList:
    """Lists inputs for a model."""

    return client.request(
        method="get", path=f"/projects/{project_uuid}/models/{model_uuid}/inputs/", response_class=InputList
    )



import time
from typing import Any, Dict, List, Optional, Union
from uuid import UUID

from myst.adapters.timing import to_timing_create
from myst.adapters.utils import drop_unset_values, get_resource_uuid
from myst.client import get_client
from myst.core.time.time import Time
from myst.models.base_model import BaseModel
from myst.models.timing import AbsoluteOrCronTiming, CronTiming, TimeRangeBoundary, from_time_range_boundary
from myst.models.types import UNSET, OptionalArgument, Unset, UUIDOrStr, to_uuid
from myst.openapi.api.projects.hpos import create_hpo, get_hpo, list_hpos, run_hpo, update_hpo
from myst.openapi.models.hpo_create import HPOCreate
from myst.openapi.models.hpo_job_state import HPOJobState
from myst.openapi.models.hpo_update import HPOUpdate
from myst.openapi.models.hyperopt_create import HyperoptCreate
from myst.openapi.models.log_uniform import LogUniform as LogUniformModel
from myst.openapi.models.q_log_uniform import QLogUniform as QLogUniformModel
from myst.openapi.models.q_uniform import QUniform as QUniformModel
from myst.openapi.models.uniform import Uniform as UniformModel
from myst.resources.hpo_job import HPOJob
from myst.resources.hpo_result import HPOResult
from myst.resources.model import Model
from myst.resources.project import Project
from myst.resources.resource import Resource


class Uniform(BaseModel):
    lower: float
    upper: float


class QUniform(BaseModel):
    lower: float
    upper: float
    q: float


class LogUniform(BaseModel):
    lower: float
    upper: float
    base: Optional[int] = 10


class QLogUniform(BaseModel):
    lower: float
    upper: float
    q: float
    base: Optional[int] = 10


class Hyperopt(BaseModel):
    num_trials: Optional[int] = 10
    max_concurrent_trials: Optional[int] = 1
    n_startup_jobs: Optional[int] = None


SearchAlgorithmCreate = Union[HyperoptCreate]


def _to_search_algorithm_create(hyper_opt: OptionalArgument[Hyperopt]) -> OptionalArgument[SearchAlgorithmCreate]:
    if isinstance(hyper_opt, Hyperopt):
        return HyperoptCreate(object="SearchAlgorithm", **hyper_opt.dict())
    else:
        return hyper_opt


def _to_search_space_create(
    search_space: OptionalArgument[Dict[str, Optional[Union[Uniform, QUniform, LogUniform, QLogUniform]]]]
) -> OptionalArgument[Dict[str, Optional[Union[UniformModel, QUniformModel, LogUniformModel, QLogUniformModel]]]]:
    if search_space is None:
        return None
    elif isinstance(search_space, Unset):
        return UNSET

    search_space_create: Dict[str, Optional[Union[UniformModel, QUniformModel, LogUniformModel, QLogUniformModel]]] = {}
    for key, value in search_space.items():
        if isinstance(value, Uniform):
            search_space_create[key] = UniformModel(object="Sampler", type="Uniform", **value.dict())
        elif isinstance(value, QUniform):
            search_space_create[key] = QUniformModel(object="Sampler", type="QUniform", **value.dict())
        elif isinstance(value, LogUniform):
            search_space_create[key] = LogUniformModel(object="Sampler", type="LogUniform", **value.dict())
        elif isinstance(value, QLogUniform):
            search_space_create[key] = QLogUniformModel(object="Sampler", type="QLogUniform", **value.dict())
        else:
            raise TypeError(
                "The `search_space` argument must be of type `Uniform`, `QUniform`, `LogUniform`, or `QLogUniform`."
            )

    return search_space_create


class HPO(Resource):

    title: str
    description: Optional[str]
    project: UUID
    model: UUID
    test_start_time: Time
    test_end_time: Time
    fit_start_timing: TimeRangeBoundary
    fit_end_timing: TimeRangeBoundary
    fit_reference_timing: AbsoluteOrCronTiming
    predict_start_timing: TimeRangeBoundary
    predict_end_timing: TimeRangeBoundary
    predict_reference_timing: CronTiming
    search_algorithm: Hyperopt
    search_space: Dict[str, Union[Uniform, QUniform, LogUniform, QLogUniform]]

    @classmethod
    def create(
        cls,
        project: Union["Project", UUIDOrStr],
        title: str,
        model: Union["Model", UUIDOrStr],
        test_start_time: Time,
        test_end_time: Time,
        fit_start_timing: TimeRangeBoundary,
        fit_end_timing: TimeRangeBoundary,
        fit_reference_timing: AbsoluteOrCronTiming,
        predict_start_timing: TimeRangeBoundary,
        predict_end_timing: TimeRangeBoundary,
        predict_reference_timing: CronTiming,
        search_algorithm: Hyperopt,
        search_space: Dict[str, Optional[Union[Uniform, QUniform, LogUniform, QLogUniform]]],
        description: Optional[str] = None,
    ) -> "HPO":
        """Creates a new HPO."""
        hpo = create_hpo.request_sync(
            client=get_client(),
            project_uuid=str(get_resource_uuid(project)),
            json_body=HPOCreate(
                object="HPO",
                title=title,
                description=description,
                model=str(get_resource_uuid(model)),
                test_start_time=test_start_time and test_start_time.to_iso_string() or None,
                test_end_time=test_end_time and test_end_time.to_iso_string() or None,
                fit_start_timing=to_timing_create(from_time_range_boundary(fit_start_timing)),
                fit_end_timing=to_timing_create(from_time_range_boundary(fit_end_timing)),
                fit_reference_timing=to_timing_create(fit_reference_timing),
                predict_start_timing=to_timing_create(from_time_range_boundary(predict_start_timing)),
                predict_end_timing=to_timing_create(from_time_range_boundary(predict_end_timing)),
                predict_reference_timing=to_timing_create(predict_reference_timing),
                search_algorithm=_to_search_algorithm_create(search_algorithm),
                search_space=_to_search_space_create(search_space),
            ),
        )
        return HPO.parse_obj(hpo.dict())

    @classmethod
    def get(cls, project: Union["Project", UUIDOrStr], uuid: UUIDOrStr) -> "HPO":
        """Gets a HPO by its identifier."""
        hpo = get_hpo.request_sync(
            client=get_client(), project_uuid=str(get_resource_uuid(project)), uuid=str(to_uuid(uuid))
        )

        return HPO.parse_obj(hpo.dict())

    def update(
        self,
        title: OptionalArgument[str] = UNSET,
        model: OptionalArgument[Union[Model, UUIDOrStr]] = UNSET,
        test_start_time: OptionalArgument[Time] = UNSET,
        test_end_time: OptionalArgument[Time] = UNSET,
        fit_start_timing: OptionalArgument[TimeRangeBoundary] = UNSET,
        fit_end_timing: OptionalArgument[TimeRangeBoundary] = UNSET,
        fit_reference_timing: OptionalArgument[AbsoluteOrCronTiming] = UNSET,
        predict_start_timing: OptionalArgument[TimeRangeBoundary] = UNSET,
        predict_end_timing: OptionalArgument[TimeRangeBoundary] = UNSET,
        predict_reference_timing: OptionalArgument[CronTiming] = UNSET,
        search_algorithm: OptionalArgument[Hyperopt] = UNSET,
        search_space: OptionalArgument[Dict[str, Optional[Union[Uniform, QUniform, LogUniform, QLogUniform]]]] = UNSET,
        description: OptionalArgument[str] = UNSET,
    ) -> "HPO":
        """Updates an HPO."""
        hpo_update: Dict[str, Any] = dict(
            object="HPO",
            title=title,
            description=description,
            model=str(get_resource_uuid(model)) if isinstance(model, Model) else model,
            test_start_time=test_start_time.to_iso_string() if isinstance(test_start_time, Time) else test_start_time,
            test_end_time=test_end_time.to_iso_string() if isinstance(test_end_time, Time) else test_end_time,
            fit_start_timing=to_timing_create(from_time_range_boundary(fit_start_timing)),
            fit_end_timing=to_timing_create(from_time_range_boundary(fit_end_timing)),
            fit_reference_timing=to_timing_create(fit_reference_timing),
            predict_start_timing=to_timing_create(from_time_range_boundary(predict_start_timing)),
            predict_end_timing=to_timing_create(from_time_range_boundary(predict_end_timing)),
            predict_reference_timing=to_timing_create(predict_reference_timing),
            search_algorithm=_to_search_algorithm_create(search_algorithm),
            search_space=_to_search_space_create(search_space),
        )

        hpo = update_hpo.request_sync(
            client=get_client(),
            project_uuid=str(self.project),
            uuid=str(self.uuid),
            json_body=HPOUpdate.parse_obj(drop_unset_values(hpo_update)),
        )

        return HPO.parse_obj(hpo.dict())

    @classmethod
    def list(cls, project: Union["Project", UUIDOrStr]) -> List["HPO"]:
        """Lists all HPOs associated with this project."""
        hpos = list_hpos.request_sync(client=get_client(), project_uuid=str(get_resource_uuid(project)))

        return [HPO.parse_obj(hpo.dict()) for hpo in hpos.data]

    @property
    def state(self) -> HPOJobState:
        """Returns the state of the HPO from the latest HPO job."""
        # Refresh the related job to get the most up to date state of the HPO.
        # Note that this will make a `get` request.
        return self.get_job().state

    def run(self) -> HPOJob:
        """Runs the HPO."""
        hpo_job = run_hpo.request_sync(client=get_client(), project_uuid=str(self.project), uuid=str(self.uuid))

        return HPOJob.parse_obj(hpo_job.dict())

    def get_job(self) -> HPOJob:
        """Gets the latest job associated with this HPO.

        Returns:
            the HPO job

        Raises:
            ValueError: if the job uuid cannot be inferred from a recent run
        """
        all_hpo_jobs = HPOJob.list(project=self.project, hpo=self)
        all_hpo_jobs.sort(key=lambda job: job.create_time, reverse=True)

        if len(all_hpo_jobs) == 0:
            raise ValueError("This HPO has not yet been `run`. Please run the HPO before requesting a job.")
        else:
            return all_hpo_jobs[0]

    def get_result(self) -> HPOResult:
        """Gets the latest result associated with a job for this HPO.

        Returns:
            the HPO result

        Raises:
            ValueError: if the job state is not `SUCCESS`
        """
        job = self.get_job()
        if job.state != "SUCCESS":
            raise (ValueError(f"Job state is `{job.state}`, should be `SUCCESS` to get result."))

        return HPOResult.get(project=self.project, hpo=self, job=job)

    def wait_until_completed(self) -> None:
        """Returns the hpo job once it has completed."""
        while True:
            refreshed_hpo_job = self.get_job()

            if refreshed_hpo_job.is_completed():
                break

            time.sleep(60)



from myst.client import Client
from myst.openapi.models.model_fit_policy_get import ModelFitPolicyGet


def request_sync(client: Client, project_uuid: str, model_uuid: str, uuid: str) -> ModelFitPolicyGet:
    """Deletes a model fit policy."""

    return client.request(
        method="delete",
        path=f"/projects/{project_uuid}/models/{model_uuid}/fit_policies/{uuid}",
        response_class=ModelFitPolicyGet,
    )



from myst.client import Client
from myst.openapi.models.project_nodes_list import ProjectNodesList


def request_sync(client: Client, project_uuid: str) -> ProjectNodesList:
    """Lists nodes for a given project."""

    return client.request(method="get", path=f"/projects/{project_uuid}/nodes/", response_class=ProjectNodesList)



from typing import Any, Dict, Optional, Type, TypeVar

import httpx
import pydantic

from myst.auth.credentials import Credentials
from myst.models.base_model import BaseModel
from myst.models.exceptions import MystAPIError, MystClientError, NotFound, UnauthenticatedError
from myst.models.http_errors import HTTPValidationError, StringDetailError
from myst.settings import settings
from myst.version import get_package_version

RequestModelType = TypeVar("RequestModelType", bound=BaseModel)
ResponseModelType = TypeVar("ResponseModelType", bound=BaseModel)


class Client:
    """HTTP client for interacting with the Myst Platform API."""

    API_VERSION = "v1alpha2"
    USER_AGENT_PREFORMAT = "Myst/{api_version} PythonBindings/{package_version}"

    def __init__(
        self, credentials: Optional[Credentials] = None, timeout: Optional[int] = None, api_host: Optional[str] = None
    ):
        """A wrapper object providing convenient API access with credentials."""
        # Construct an httpx client so that we can reuse a single TCP connection and configuration.
        self.base_url = (api_host or settings.MYST_API_HOST).rstrip("/")
        self._client = httpx.Client(
            timeout=timeout or settings.API_TIMEOUT_SEC,
            headers={"User-Agent": self.user_agent},
            base_url=f"{self.base_url}/{self.API_VERSION}",
        )
        self._credentials = credentials

    def __del__(self) -> None:
        """Cleans up underlying client resources."""
        try:
            self._client.close()
        except Exception:
            # We aren't too concerned with issues that occur while cleaning up connections.
            pass

    def authenticate(self, credentials: Credentials) -> None:
        """Authenticates this client using the given credentials."""
        self._credentials = credentials

    @property
    def user_agent(self) -> str:
        """Gets the `User-Agent` header string to send to the Myst API."""
        # Infer Myst API version and Myst Python client library version for current active versions.
        return self.USER_AGENT_PREFORMAT.format(api_version=self.API_VERSION, package_version=get_package_version())

    def _get_credentials(self) -> Credentials:
        if self._credentials is None:
            raise UnauthenticatedError("No client credentials provided.")

        return self._credentials

    def request(
        self,
        method: str,
        path: str,
        response_class: Type[ResponseModelType],
        params: Optional[Dict[str, Any]] = None,
        request_model: Optional[RequestModelType] = None,
    ) -> ResponseModelType:
        """Executes a request for the given HTTP method and URL, handling JSON serialization and deserialization.

        Args:
            method: HTTP verb to execute, e.g. "GET", "POST", etc.
            path: path relative to the base URL, e.g. "/time_series/"
            response_class: name of the model class to parse the response content into
            params: HTTP query parameters
            request_model: JSON-able model instance to pass in request body

        Raises:
            MystClientError: client error (HTTP 4xx), including further details in the case of 422
            MystAPIError: server error (HTTP 500) or unrecognized error
            NotFound: client error (HTTP 404)

        Returns:
            parsed and validated instance of indicated response class
        """
        # Convert the Pydantic request model into JSON form.
        content = request_model.json(exclude_unset=True, by_alias=True) if request_model else None

        # Make the API request.
        response = self._client.request(
            method=method,
            url=path,
            params=params,
            headers={"content-type": "application/json", "Authorization": f"Bearer {self._get_credentials().token}"},
            content=content,
        )

        # Handle the httpx response.
        if response.status_code in (200, 201):
            return response_class.parse_raw(response.content)
        elif response.status_code == 422:
            # Special-case 422 validation errors.
            try:
                raise MystClientError(HTTPValidationError.parse_raw(response.content))
            except pydantic.ValidationError:
                raise MystClientError(StringDetailError.parse_raw(response.content))
        elif response.status_code == 404:
            raise NotFound(message=StringDetailError.parse_raw(response.content).detail)
        elif 400 <= response.status_code < 500:
            # All other client errors.
            raise MystClientError(response.json())
        elif response.status_code == 500:
            raise MystAPIError(status_code=response.status_code, message="Internal server error")
        else:
            raise MystAPIError(
                message=f"An unexpected error occurred: {response.content}", status_code=response.status_code
            )


# Lazily instantiate the global client so that we don't form a connection pool at import time.
_client: Optional[Client] = None


def get_client() -> Client:
    """Returns the global Myst client instance."""
    global _client

    if _client is None:
        _client = Client()

    return _client


def set_client(client: Client) -> None:
    """Sets the global Myst client instance.

    Useful for e.g. configuring timeouts globally.

    Args:
        client: the client to use
    """
    global _client

    _client = client



from myst.client import Client
from myst.openapi.models.user_get import UserGet


def request_sync(client: Client, uuid: str) -> UserGet:
    """Gets the requested user."""

    return client.request(method="get", path=f"/users/{uuid}", response_class=UserGet)



from typing import Optional, Union

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.absolute_timing_get import AbsoluteTimingGet
from myst.openapi.models.cron_timing_get import CronTimingGet
from myst.openapi.models.hyperopt_get import HyperoptGet
from myst.openapi.models.relative_timing_get import RelativeTimingGet
from myst.openapi.models.search_space import SearchSpace


class HPOGet(base_model.BaseModel):
    """HPO schema for get responses."""

    object_: Literal["HPO"] = Field(..., alias="object")
    uuid: str
    create_time: str
    title: str
    project: str
    model: str
    creator: str
    search_space: SearchSpace
    search_algorithm: HyperoptGet
    fit_start_timing: Union[AbsoluteTimingGet, RelativeTimingGet]
    fit_end_timing: Union[AbsoluteTimingGet, RelativeTimingGet]
    predict_start_timing: Union[AbsoluteTimingGet, RelativeTimingGet]
    predict_end_timing: Union[AbsoluteTimingGet, RelativeTimingGet]
    update_time: Optional[str] = None
    description: Optional[str] = None
    test_start_time: Optional[str] = None
    test_end_time: Optional[str] = None
    fit_reference_timing: Optional[Union[AbsoluteTimingGet, CronTimingGet]] = None
    predict_reference_timing: Optional[CronTimingGet] = None



from myst.client import Client
from myst.openapi.models.source_create import SourceCreate
from myst.openapi.models.source_get import SourceGet


def request_sync(client: Client, project_uuid: str, json_body: SourceCreate) -> SourceGet:
    """Creates a source."""

    return client.request(
        method="post", path=f"/projects/{project_uuid}/sources/", response_class=SourceGet, request_model=json_body
    )



from typing import Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model


class TimeSeriesInsertResultGet(base_model.BaseModel):
    """Time series insert result schema for get responses."""

    object_: Literal["NodeResult"] = Field(..., alias="object")
    uuid: str
    create_time: str
    type: Literal["TimeSeriesInsertResult"]
    node: str
    update_time: Optional[str] = None



import enum
from dataclasses import dataclass
from typing import Optional
from uuid import UUID

from myst.connectors.operation_connector import OperationConnector
from myst.core.time.time_delta import TimeDelta


@enum.unique
class GroupName(str, enum.Enum):
    OPERANDS = "operands"


@enum.unique
class AggregationFunction(str, enum.Enum):
    COUNT = "count"
    SUM = "sum"
    MEAN = "mean"
    MEDIAN = "median"
    MIN = "min"
    MAX = "max"


@dataclass
class RollingWindowParameters:
    window_period: TimeDelta
    min_periods: int
    centered: bool
    aggregation_function: AggregationFunction


@dataclass
class DifferencingParameters:
    differencing_period: Optional[TimeDelta] = None
    differencing_order: Optional[int] = None


@dataclass
class ShiftParameters:
    shift_period: Optional[TimeDelta] = None


class TimeTransformations(OperationConnector):
    def __init__(
        self,
        rolling_window_parameters: Optional[RollingWindowParameters] = None,
        differencing_parameters: Optional[DifferencingParameters] = None,
        shift_parameters: Optional[ShiftParameters] = None,
    ) -> None:
        super().__init__(
            uuid=UUID("e4bebd6d-d9d2-4103-8c94-95fa162eddc0"),
            parameters=dict(
                rolling_window_parameters=rolling_window_parameters,
                differencing_parameters=differencing_parameters,
                shift_parameters=shift_parameters,
            ),
        )



from myst.client import Client
from myst.openapi.models.project_node_result_list import ProjectNodeResultList


def request_sync(client: Client, project_uuid: str) -> ProjectNodeResultList:
    """Lists results."""

    return client.request(
        method="get", path=f"/projects/{project_uuid}/nodes/-/results/", response_class=ProjectNodeResultList
    )



from myst.models import base_model
from myst.openapi.models.operation import Operation


class ShareableResourceShare(base_model.BaseModel):
    """Schema for share requests."""

    accessor: str
    access_level: Operation



"""The official `myst` python package owned by Myst AI, Inc."""

from myst.auth.credentials import authenticate, authenticate_with_service_account
from myst.client import Client, get_client, set_client
from myst.core.data.time_array import TimeArray
from myst.core.time.time import Time
from myst.core.time.time_delta import TimeDelta
from myst.models.timing import AbsoluteTiming, CronTiming, RelativeTiming
from myst.resources import hpo
from myst.resources.backtest import Backtest, BacktestJob
from myst.resources.hpo import HPO
from myst.resources.input import ModelInput, OperationInput
from myst.resources.layer import TimeSeriesLayer
from myst.resources.model import Model
from myst.resources.operation import Operation
from myst.resources.policy import ModelFitPolicy, Policy, TimeSeriesRunPolicy
from myst.resources.project import Project
from myst.resources.source import Source
from myst.resources.time_series import TimeSeries
from myst.settings import settings
from myst.version import get_package_version

__version__ = get_package_version()


__all__ = [
    "AbsoluteTiming",
    "authenticate_with_service_account",
    "authenticate",
    "Client",
    "get_client",
    "set_client",
    "Backtest",
    "HPO",
    "CronTiming",
    "ModelInput",
    "OperationInput",
    "TimeSeriesLayer",
    "Model",
    "ModelFitPolicy",
    "Operation",
    "Policy",
    "Project",
    "RelativeTiming",
    "settings",
    "Source",
    "Time",
    "TimeArray",
    "TimeDelta",
    "TimeSeries",
    "TimeSeriesRunPolicy",
    "hpo",
]



from typing import TYPE_CHECKING, List, Union
from uuid import UUID

from myst.adapters.utils import get_resource_uuid
from myst.client import get_client
from myst.models.types import UUIDOrStr
from myst.openapi.api.projects.hpos.results import get_hpo_result
from myst.resources.hpo_trial import HPOTrial
from myst.resources.resource import Resource

if TYPE_CHECKING:  # Avoid circular imports.
    from myst.resources.hpo import HPO
    from myst.resources.hpo_job import HPOJob
    from myst.resources.project import Project


class HPOResult(Resource):

    uuid: UUID
    best_trial: HPOTrial
    trials: List[HPOTrial]

    @classmethod
    def get(
        cls, project: Union["Project", UUIDOrStr], hpo: Union["HPO", UUIDOrStr], job: Union["HPOJob", UUIDOrStr]
    ) -> "HPOResult":
        hpo_result_list = get_hpo_result.request_sync(
            client=get_client(),
            project_uuid=str(get_resource_uuid(project)),
            hpo_uuid=str(get_resource_uuid(hpo)),
            job_uuid=str(get_resource_uuid(job)),
        )

        return HPOResult.parse_obj(hpo_result_list.data[0].dict())



from typing import Any, Dict
from uuid import UUID


class Connector:
    """Base class for connectors."""

    def __init__(self, uuid: UUID, parameters: Dict[str, Any]) -> None:
        self._uuid = uuid
        self._parameters = parameters

    @property
    def uuid(self) -> UUID:
        return self._uuid

    @property
    def parameters(self) -> Dict[str, Any]:
        return self._parameters

    def parameters_exclude_none(self) -> Dict[str, Any]:
        return {k: v for k, v in self.parameters.items() if v is not None}



import time
from typing import Any, Dict, List, Optional, Union
from uuid import UUID

from myst.adapters.timing import to_timing_create
from myst.adapters.utils import drop_unset_values, get_resource_uuid
from myst.client import get_client
from myst.core.time.time import Time
from myst.models.timing import AbsoluteOrCronTiming, CronTiming, TimeRangeBoundary, from_time_range_boundary
from myst.models.types import UNSET, OptionalArgument, UUIDOrStr, to_uuid
from myst.openapi.api.projects.backtests import (
    create_backtest,
    get_backtest,
    list_backtests,
    run_backtest,
    update_backtest,
)
from myst.openapi.models.backtest_create import BacktestCreate
from myst.openapi.models.backtest_update import BacktestUpdate
from myst.openapi.models.job_state import JobState
from myst.resources.backtest_job import BacktestJob
from myst.resources.backtest_result import BacktestResult
from myst.resources.model import Model
from myst.resources.project import Project
from myst.resources.resource import Resource


class Backtest(Resource):

    title: str
    description: Optional[str]
    project: UUID
    model: UUID
    test_start_time: Time
    test_end_time: Time
    fit_start_timing: TimeRangeBoundary
    fit_end_timing: TimeRangeBoundary
    fit_reference_timing: AbsoluteOrCronTiming
    predict_start_timing: TimeRangeBoundary
    predict_end_timing: TimeRangeBoundary
    predict_reference_timing: CronTiming

    @classmethod
    def create(
        cls,
        project: Union[Project, UUIDOrStr],
        title: str,
        model: Union[Model, UUIDOrStr],
        test_start_time: Time,
        test_end_time: Time,
        fit_start_timing: TimeRangeBoundary,
        fit_end_timing: TimeRangeBoundary,
        fit_reference_timing: AbsoluteOrCronTiming,
        predict_start_timing: TimeRangeBoundary,
        predict_end_timing: TimeRangeBoundary,
        predict_reference_timing: CronTiming,
        description: Optional[str] = None,
    ) -> "Backtest":
        """Creates a new backtest."""
        backtest = create_backtest.request_sync(
            client=get_client(),
            project_uuid=str(get_resource_uuid(project)),
            json_body=BacktestCreate(
                object="Backtest",
                title=title,
                description=description,
                model=str(get_resource_uuid(model)),
                test_start_time=test_start_time.to_iso_string(),
                test_end_time=test_end_time.to_iso_string(),
                fit_start_timing=to_timing_create(from_time_range_boundary(fit_start_timing)),
                fit_end_timing=to_timing_create(from_time_range_boundary(fit_end_timing)),
                fit_reference_timing=to_timing_create(fit_reference_timing),
                predict_start_timing=to_timing_create(from_time_range_boundary(predict_start_timing)),
                predict_end_timing=to_timing_create(from_time_range_boundary(predict_end_timing)),
                predict_reference_timing=to_timing_create(predict_reference_timing),
            ),
        )

        return Backtest.parse_obj(backtest.dict())

    @classmethod
    def get(cls, project: Union[Project, UUIDOrStr], uuid: UUIDOrStr) -> "Backtest":
        """Gets a backtest by its identifier."""
        backtest = get_backtest.request_sync(
            client=get_client(), project_uuid=str(get_resource_uuid(project)), uuid=str(to_uuid(uuid))
        )

        return Backtest.parse_obj(backtest.dict())

    def update(
        self,
        title: OptionalArgument[str] = UNSET,
        model: OptionalArgument[Union[Model, UUIDOrStr]] = UNSET,
        test_start_time: OptionalArgument[Time] = UNSET,
        test_end_time: OptionalArgument[Time] = UNSET,
        fit_start_timing: OptionalArgument[TimeRangeBoundary] = UNSET,
        fit_end_timing: OptionalArgument[TimeRangeBoundary] = UNSET,
        fit_reference_timing: OptionalArgument[AbsoluteOrCronTiming] = UNSET,
        predict_start_timing: OptionalArgument[TimeRangeBoundary] = UNSET,
        predict_end_timing: OptionalArgument[TimeRangeBoundary] = UNSET,
        predict_reference_timing: OptionalArgument[CronTiming] = UNSET,
        description: OptionalArgument[str] = UNSET,
    ) -> "Backtest":
        """Updates a new backtest."""
        backtest_update: Dict[str, Any] = dict(
            object="Backtest",
            title=title,
            description=description,
            model=str(get_resource_uuid(model)) if isinstance(model, Model) else model,
            test_start_time=test_start_time.to_iso_string() if isinstance(test_start_time, Time) else test_start_time,
            test_end_time=test_end_time.to_iso_string() if isinstance(test_end_time, Time) else test_end_time,
            fit_start_timing=to_timing_create(from_time_range_boundary(fit_start_timing)),
            fit_end_timing=to_timing_create(from_time_range_boundary(fit_end_timing)),
            fit_reference_timing=to_timing_create(fit_reference_timing),
            predict_start_timing=to_timing_create(from_time_range_boundary(predict_start_timing)),
            predict_end_timing=to_timing_create(from_time_range_boundary(predict_end_timing)),
            predict_reference_timing=to_timing_create(predict_reference_timing),
        )

        backtest = update_backtest.request_sync(
            client=get_client(),
            project_uuid=str(self.project),
            uuid=str(self.uuid),
            json_body=BacktestUpdate.parse_obj(drop_unset_values(backtest_update)),
        )

        return Backtest.parse_obj(backtest.dict())

    @classmethod
    def list(cls, project: Union[Project, UUIDOrStr]) -> List["Backtest"]:
        """Lists all backtests associated with this project."""
        backtests = list_backtests.request_sync(client=get_client(), project_uuid=str(get_resource_uuid(project)))

        return [Backtest.parse_obj(backtest) for backtest in backtests.data]

    @property
    def state(self) -> JobState:
        """Returns the state of the backtest from the latest backtest job."""
        # Refresh the related job to get the most up to date state of the backtest.
        # Note that this will make a `get` request.
        return self.get_job().state

    def run(self) -> BacktestJob:
        """Runs the backtest."""
        backtest_job = run_backtest.request_sync(
            client=get_client(), project_uuid=str(self.project), uuid=str(self.uuid)
        )

        return BacktestJob.parse_obj(backtest_job.dict())

    def get_job(self) -> BacktestJob:
        """Gets the latest job associated with this backtest.

        Returns:
            the backtest job

        Raises:
            ValueError: if the job uuid cannot be inferred from a recent run
        """
        all_backtest_jobs = BacktestJob.list(project=self.project, backtest=self.uuid)

        if len(all_backtest_jobs) == 0:
            raise ValueError("This backtest has not yet been `run`. Please run the backtest before requesting a job.")
        else:
            return all_backtest_jobs[0]

    def get_result(self) -> BacktestResult:
        """Gets the latest result associated with a job for this backtest.

        Returns:
            the backtest result

        Raises:
            ValueError: if the job uuid cannot be inferred from a recent run
        """
        all_backtest_jobs = BacktestJob.list(project=self.project, backtest=self.uuid)

        if len(all_backtest_jobs) == 0:
            raise ValueError(
                "This backtest has not yet been `run`. Please run the backtest before requesting a result."
            )
        else:
            job_uuid = all_backtest_jobs[0].uuid

        return BacktestResult.get(project=self.project, backtest=self.uuid, job=job_uuid)

    def wait_until_completed(self) -> None:
        """Returns the backtest job once it has completed."""
        while True:
            refreshed_backtest_job = self.get_job()

            if refreshed_backtest_job.is_completed():
                break

            time.sleep(60)



import enum
from uuid import UUID

from myst.connectors.operation_connector import OperationConnector


@enum.unique
class GroupName(str, enum.Enum):
    LOC = "loc"
    SCALE = "scale"


@enum.unique
class Function(str, enum.Enum):
    CDF = "CDF"
    PDF = "PDF"
    PPF = "PPF"


class CauchyDistribution(OperationConnector):
    def __init__(self, function: Function, value: float) -> None:
        super().__init__(
            uuid=UUID("babb2dbf-e616-422f-aa76-14c6e60d42ae"), parameters=dict(function=function, value=value)
        )



from myst.client import Client
from myst.openapi.models.time_series_run_policy_list import TimeSeriesRunPolicyList


def request_sync(client: Client, project_uuid: str, time_series_uuid: str) -> TimeSeriesRunPolicyList:
    """Lists run policies for a time_series."""

    return client.request(
        method="get",
        path=f"/projects/{project_uuid}/time_series/{time_series_uuid}/run_policies/",
        response_class=TimeSeriesRunPolicyList,
    )



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from myst.adapters.utils import drop_unset_values, get_resource_uuid
from myst.client import get_client
from myst.connectors.operation_connector import OperationConnector
from myst.models.types import UNSET, ItemOrSlice, OptionalArgument, UUIDOrStr, to_uuid
from myst.openapi.api.projects.operations import create_operation, get_operation, update_operation
from myst.openapi.models.operation_create import OperationCreate
from myst.openapi.models.operation_update import OperationUpdate
from myst.resources.connector_node import ConnectorNode
from myst.resources.input import OperationInput
from myst.resources.time_series import TimeSeries

if TYPE_CHECKING:  # Avoid circular imports.
    from myst.resources.project import Project


class Operation(ConnectorNode):
    """A node that performs a specified transformation on its input.

    In contrast to a model, an operation has no training phase and can only be run.
    """

    @classmethod
    def create(
        cls,
        project: Union["Project", UUIDOrStr],
        title: str,
        connector: OperationConnector,
        description: Optional[str] = None,
    ) -> "Operation":
        """Creates a new operation node.

        Args:
            project: the project in which to create the operation
            title: the title of the operation
            connector: the operation connector to use in the operation node
            description: a brief description of the operation

        Returns:
            the newly created operation
        """
        operation = create_operation.request_sync(
            client=get_client(),
            project_uuid=str(get_resource_uuid(project)),
            json_body=OperationCreate(
                object="Node",
                type="Operation",
                title=title,
                description=description,
                connector_uuid=str(connector.uuid),
                parameters=connector.parameters_exclude_none(),
            ),
        )

        return Operation.parse_obj(operation.dict())

    @classmethod
    def get(cls, project: Union["Project", UUIDOrStr], uuid: UUIDOrStr) -> "Operation":
        """Gets a specific operation by its identifier."""
        operation = get_operation.request_sync(
            client=get_client(), project_uuid=str(get_resource_uuid(project)), uuid=str(to_uuid(uuid))
        )

        return Operation.parse_obj(operation.dict())

    @classmethod
    def list(cls, project: Union["Project", UUIDOrStr]) -> List["Operation"]:
        """Gets all operations by project."""
        raise NotImplementedError()

    def update(
        self,
        title: OptionalArgument[str] = UNSET,
        connector: OptionalArgument[OperationConnector] = UNSET,
        description: OptionalArgument[str] = UNSET,
    ) -> "Operation":
        """Updates operation."""
        parameters = connector.parameters_exclude_none() if isinstance(connector, OperationConnector) else UNSET
        operation_update: Dict[str, Any] = dict(
            object="Node",
            type="Operation",
            title=title,
            connector_uuid=str(connector.uuid) if isinstance(connector, OperationConnector) else UNSET,
            parameters=parameters,
            description=description,
        )

        operation = update_operation.request_sync(
            client=get_client(),
            project_uuid=str(self.project),
            uuid=str(self.uuid),
            json_body=OperationUpdate.parse_obj(drop_unset_values(operation_update)),
        )

        return Operation.parse_obj(operation.dict())

    def create_input(
        self,
        time_series: Union[TimeSeries, UUIDOrStr],
        group_name: str,
        output_index: int = 0,
        label_indexer: Optional[ItemOrSlice] = None,
    ) -> OperationInput:
        """Creates an input into this operation.

        Args:
            time_series: the time series to feed into this operation
            group_name: the name of the input group on this operation's connector to which to pass the data from this
                input
            output_index: which time dataset, out of the sequence of upstream time datasets, to pass to this operation
            label_indexer: the slice of the upstream data to pass to this operation

        Returns:
            the newly created input
        """
        return OperationInput.create(
            project=self.project,
            operation=self.uuid,
            time_series=time_series,
            group_name=group_name,
            output_index=output_index,
            label_indexer=label_indexer,
        )

    def list_inputs(self) -> List[OperationInput]:
        """Lists all inputs into this operation."""
        return OperationInput.list(project=self.project, operation=self.uuid)



from typing import List

from myst.models import base_model
from myst.openapi.models.project_get import ProjectGet


class ProjectList(base_model.BaseModel):
    """Schema for project list responses."""

    data: List[ProjectGet]
    has_more: bool



from myst.client import Client
from myst.openapi.models.model_fit_job_get import ModelFitJobGet


def request_sync(client: Client, project_uuid: str, model_uuid: str, uuid: str) -> ModelFitJobGet:
    """Gets a model fit job by its unique identifier."""

    return client.request(
        method="get",
        path=f"/projects/{project_uuid}/models/{model_uuid}/fit_jobs/{uuid}",
        response_class=ModelFitJobGet,
    )



from myst.client import Client
from myst.openapi.models.deployment_list import DeploymentList


def request_sync(client: Client, project_uuid: str) -> DeploymentList:
    """Lists deployments for a given project."""

    return client.request(method="get", path=f"/projects/{project_uuid}/deployments/", response_class=DeploymentList)



from typing import Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.source_connector_get_parameters_schema import SourceConnectorGetParametersSchema


class SourceConnectorGet(base_model.BaseModel):
    """Schema for source connector get responses."""

    object_: Literal["Connector"] = Field(..., alias="object")
    type: Literal["SourceConnector"]
    uuid: str
    title: str
    provider: str
    description: str
    parameters_schema: SourceConnectorGetParametersSchema
    icon_url: Optional[str] = None



from typing import Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model


class DeploymentGet(base_model.BaseModel):
    """Schema for deployment get responses."""

    object_: Literal["Deployment"] = Field(..., alias="object")
    uuid: str
    create_time: str
    title: str
    creator: str
    update_time: Optional[str] = None
    activate_time: Optional[str] = None
    deactivate_time: Optional[str] = None



from myst.client import Client
from myst.openapi.models.model_get import ModelGet


def request_sync(client: Client, project_uuid: str, uuid: str) -> ModelGet:
    """Gets a model by its unique identifier."""

    return client.request(method="get", path=f"/projects/{project_uuid}/models/{uuid}", response_class=ModelGet)



from typing import List

from myst.models import base_model
from myst.openapi.models.user_get import UserGet


class UserList(base_model.BaseModel):
    """Schema for user list responses."""

    data: List[UserGet]
    has_more: bool



from myst.client import Client
from myst.openapi.models.model_fit_policy_get import ModelFitPolicyGet
from myst.openapi.models.model_fit_policy_update import ModelFitPolicyUpdate


def request_sync(
    client: Client, project_uuid: str, model_uuid: str, uuid: str, json_body: ModelFitPolicyUpdate
) -> ModelFitPolicyGet:
    """Updates a model fit policy."""

    return client.request(
        method="patch",
        path=f"/projects/{project_uuid}/models/{model_uuid}/fit_policies/{uuid}",
        response_class=ModelFitPolicyGet,
        request_model=json_body,
    )



import enum
from typing import Optional
from uuid import UUID

from myst.connectors.model_connector import ModelConnector


@enum.unique
class GroupName(str, enum.Enum):
    FEATURES = "features"
    TARGETS = "targets"


class ElasticNet(ModelConnector):
    def __init__(
        self,
        alpha: Optional[float] = None,
        l1_ratio: Optional[float] = None,
        fit_intercept: Optional[bool] = None,
        max_iter: Optional[int] = None,
    ) -> None:
        super().__init__(
            uuid=UUID("f2aaf05d-4fb4-4106-979e-a1ebd36efb2b"),
            parameters=dict(alpha=alpha, l1_ratio=l1_ratio, fit_intercept=fit_intercept, max_iter=max_iter),
        )



from typing import Optional, Union

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.absolute_timing_create import AbsoluteTimingCreate
from myst.openapi.models.cron_timing_create import CronTimingCreate
from myst.openapi.models.relative_timing_create import RelativeTimingCreate


class ModelFitPolicyCreate(base_model.BaseModel):
    """Schema for model fit policy create requests."""

    object_: Literal["Policy"] = Field(..., alias="object")
    schedule_timing: Union[AbsoluteTimingCreate, RelativeTimingCreate, CronTimingCreate]
    start_timing: Union[AbsoluteTimingCreate, RelativeTimingCreate]
    end_timing: Union[AbsoluteTimingCreate, RelativeTimingCreate]
    type: Optional[Literal["ModelFitPolicy"]] = "ModelFitPolicy"
    active: Optional[bool] = True



from myst.client import Client
from myst.openapi.models.user_create import UserCreate
from myst.openapi.models.user_get import UserGet


def request_sync(client: Client, json_body: UserCreate) -> UserGet:
    """Creates a user."""

    return client.request(method="post", path=f"/users/", response_class=UserGet, request_model=json_body)



from typing import Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model


class RelativeTimingCreate(base_model.BaseModel):
    """Relative timing schema for create requests."""

    type: Literal["RelativeTiming"]
    object_: Optional[Literal["Timing"]] = Field("Timing", alias="object")
    frequency: Optional[str] = None
    offset: Optional[str] = None
    time_zone: Optional[str] = None



import enum
from uuid import UUID

from myst.connectors.operation_connector import OperationConnector
from myst.core.time.time_delta import TimeDelta


@enum.unique
class GroupName(str, enum.Enum):
    OPERANDS = "operands"


@enum.unique
class ResamplingFunction(str, enum.Enum):
    INTERPOLATE = "interpolate"
    PAD = "pad"
    FIRST = "first"
    LAST = "last"
    MEAN = "mean"
    MEDIAN = "median"
    MIN = "min"
    MAX = "max"
    SUM = "sum"


class Resampling(OperationConnector):
    def __init__(self, sample_period: TimeDelta, resampling_function: ResamplingFunction) -> None:
        super().__init__(
            uuid=UUID("7d6d92d6-5fc2-47b7-9f3c-7ae73cf30b7f"),
            parameters=dict(sample_period=sample_period, resampling_function=resampling_function),
        )



import enum
from dataclasses import dataclass
from typing import List, Optional
from uuid import UUID

from myst.connectors.source_connector import SourceConnector


@enum.unique
class YesEnergyAggregation(str, enum.Enum):
    AVG = "AVG"
    MAX = "MAX"
    MIN = "MIN"


@dataclass
class YesEnergyItem:
    datatype: str
    object_id: int
    forecast_vintage_offset: Optional[str] = None


class YesEnergy(SourceConnector):
    def __init__(self, items: List[YesEnergyItem], stat: YesEnergyAggregation = YesEnergyAggregation.AVG) -> None:
        super().__init__(uuid=UUID("207da43c-d284-44ee-90d0-61b96fa4df1c"), parameters=dict(items=items, stat=stat))



from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model


class Uniform(base_model.BaseModel):
    """"""

    object_: Literal["Sampler"] = Field(..., alias="object")
    type: Literal["Uniform"]
    lower: float
    upper: float



from myst.client import Client
from myst.openapi.models.project_create import ProjectCreate
from myst.openapi.models.project_get import ProjectGet


def request_sync(client: Client, json_body: ProjectCreate) -> ProjectGet:
    """Creates a project."""

    return client.request(method="post", path=f"/projects/", response_class=ProjectGet, request_model=json_body)



from typing import List

from myst.models import base_model
from myst.openapi.models.layer_get import LayerGet


class LayerList(base_model.BaseModel):
    """Schema for layer list responses."""

    data: List[LayerGet]
    has_more: bool



from typing import Optional, Union

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.absolute_timing_create import AbsoluteTimingCreate
from myst.openapi.models.cron_timing_create import CronTimingCreate
from myst.openapi.models.relative_timing_create import RelativeTimingCreate


class ModelFitPolicyUpdate(base_model.BaseModel):
    """Schema for model fit policy update requests."""

    object_: Optional[Literal["Policy"]] = Field(..., alias="object")
    uuid: Optional[str] = None
    create_time: Optional[str] = None
    update_time: Optional[str] = None
    type: Optional[Literal["ModelFitPolicy"]] = None
    creator: Optional[str] = None
    schedule_timing: Optional[Union[AbsoluteTimingCreate, RelativeTimingCreate, CronTimingCreate]] = None
    active: Optional[bool] = None
    node: Optional[str] = None
    start_timing: Optional[Union[AbsoluteTimingCreate, RelativeTimingCreate]] = None
    end_timing: Optional[Union[AbsoluteTimingCreate, RelativeTimingCreate]] = None



from enum import Enum


class ConnectorGetType(str, Enum):
    SOURCECONNECTOR = "SourceConnector"
    OPERATIONCONNECTOR = "OperationConnector"
    MODELCONNECTOR = "ModelConnector"

    def __str__(self) -> str:
        return str(self.value)



from enum import Enum


class HPOJobState(str, Enum):
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"

    def __str__(self) -> str:
        return str(self.value)



from datetime import datetime
from typing import Any, Generator, Union

import pandas as pd

from myst.core.time.time_delta import TimeDelta


class Time:
    """Container for time zone-aware time objects that can be mapped from and to various third-party time objects."""

    def __init__(self, value: Union["Time", datetime, str]) -> None:
        """A `Time` container is initialized with a datetime or an ISO 8601 formatted string.

        The full ISO 8601 format specification is `YYYY-MM-DDThh:mm:ssZ`. The `Z` suffix indicates the UTC time zone.

        Note that we only support ISO strings that are in UTC, to avoid possible ambiguity around daylight savings.

        Args:
            value: ISO 8601 formatted string in UTC, or time zone-aware datetime
        """
        self._time: pd.Timestamp

        if isinstance(value, Time):
            self._time = value._time
        elif isinstance(value, datetime):
            pandas_timestamp = pd.Timestamp(value)

            if pandas_timestamp.tz is None:
                raise ValueError("Input timestamp must be time zone-aware.")

            self._time = pandas_timestamp.tz_convert(tz="UTC")
        elif isinstance(value, str):
            if not value.endswith("Z"):
                raise ValueError("The UTC time zone in passed ISO string must be indicated by the `Z` suffix.")

            pandas_timestamp = pd.Timestamp(value)

            if pandas_timestamp.tz.zone != "UTC":
                raise ValueError("The time zone in passed ISO string must be UTC, because it is ambiguous otherwise.")

            self._time = pandas_timestamp
        else:
            raise TypeError("Argument to `Time` constructor must be `Time`, datetime, or str.")

    def __eq__(self, other: object) -> bool:
        """Returns True if this time and other represent the same instant in time, False otherwise."""
        return isinstance(other, Time) and self._time == other._time

    def __ne__(self, other: object) -> bool:
        """Returns True if this time and other do not represent the same instant in time, False otherwise."""
        return not self.__eq__(other)

    def __add__(self, other: TimeDelta) -> "Time":
        """Returns a new `Time` resulting from the addition of this object and the provided `TimeDelta`."""
        if not isinstance(other, TimeDelta):
            raise ValueError("Only a `TimeDelta` instance may be added to a `Time`.")

        return Time(pd.Timestamp(self.to_pandas_timestamp() + other.to_pandas_date_offset()))

    def __radd__(self, other: Any) -> "Time":
        """Returns a new `Time` resulting from the addition of this object and the provided `TimeDelta`."""
        return self.__add__(other)

    def __sub__(self, other: TimeDelta) -> "Time":
        """Returns a new `Time` resulting from the subtraction of the provided `TimeDelta` from this object."""
        if not isinstance(other, TimeDelta):
            raise ValueError("Only a `TimeDelta` instance may be subtracted from a `Time`.")

        return Time(pd.Timestamp(self.to_pandas_timestamp() - other.to_pandas_date_offset()))

    def __lt__(self, other: "Time") -> bool:
        """Returns True if this time is less than the passed time, False otherwise."""
        return self._time < other._time

    def __le__(self, other: "Time") -> bool:
        """Returns True if this time is less than or equal to the passed time, False otherwise."""
        return self._time <= other._time

    def __gt__(self, other: "Time") -> bool:
        """Returns True if this time is greater than the passed time, False otherwise."""
        return self._time > other._time

    def __ge__(self, other: "Time") -> bool:
        """Returns True if this time is greater than or equal to the passed time, False otherwise."""
        return self._time >= other._time

    def __repr__(self) -> str:
        """Returns an encapsulated ISO 8601 formatted string representation for this time."""
        return f"<Time: {self.to_iso_string()}>"

    def __str__(self) -> str:
        """Returns the ISO 8601 formatted string representation for this time."""
        return self.to_iso_string()

    def __hash__(self) -> int:
        """Returns a hash value for this time."""
        return hash(self._time)

    def to_iso_string(self) -> str:
        """Returns the ISO 8601 formatted string representation of this time."""
        return self._time.isoformat().replace("+00:00", "Z")

    def to_pandas_timestamp(self) -> pd.Timestamp:
        """Returns the `pandas.Timestamp` equivalent of this time."""
        return self._time

    def to_datetime(self) -> datetime:
        """Returns the `datetime.datetime` equivalent of this time."""
        return self._time.to_pydatetime()

    @classmethod
    def __get_validators__(cls) -> Generator:
        yield cls.validate

    @classmethod
    def validate(cls, v: Any) -> "Time":
        return cls(v)



from myst.client import Client
from myst.openapi.models.backtest_job_get import BacktestJobGet


def request_sync(client: Client, project_uuid: str, uuid: str) -> BacktestJobGet:
    """Runs jobs for a backtest."""

    return client.request(
        method="post", path=f"/projects/{project_uuid}/backtests/{uuid}:run", response_class=BacktestJobGet
    )



import enum
from dataclasses import dataclass
from typing import Mapping, MutableMapping, Optional, Tuple, Union
from uuid import UUID

from myst.connectors.source_connectors import cleaned_observations, enhanced_forecast, historical_hourly_conditions
from myst.core.time.time_delta import TimeDelta
from myst.recipes.metar_station import MetarStation
from myst.recipes.time_series_recipe import TimeSeriesRecipe
from myst.resources.layer import TimeSeriesLayer
from myst.resources.project import Project
from myst.resources.source import Source
from myst.resources.time_series import TimeSeries

_HISTORICAL_APIS_CUTOVER_BOUNDARY = TimeDelta("-PT23H")
_HISTORICAL_TO_FORECAST_CUTOVER_BOUNDARY = TimeDelta("PT1H")
_THE_WEATHER_COMPANY_DATA_SAMPLE_PERIOD = TimeDelta("PT1H")


@dataclass
class _TWCFieldTriple:
    cleaned_observations_field: cleaned_observations.Field
    historical_hourly_conditions_field: historical_hourly_conditions.Field
    enhanced_forecast_field: enhanced_forecast.Field


@enum.unique
class Field(str, enum.Enum):
    """Data fields available from The Weather Company APIs involved in this recipe."""

    CLOUD_COVERAGE = "Cloud Coverage"
    DEW_POINT_TEMPERATURE = "Dew Point Temperature"
    RELATIVE_HUMIDITY = "Relative Humidity"
    TEMPERATURE = "Temperature"
    WIND_CHILL_TEMPERATURE = "Wind Chill Temperature"
    WIND_DIRECTION = "Wind Direction"
    WIND_SPEED = "Wind Speed"


_FIELD_MAPPINGS: Mapping[Field, _TWCFieldTriple] = {
    Field.CLOUD_COVERAGE: _TWCFieldTriple(
        cleaned_observations_field=cleaned_observations.Field.CLOUD_COVERAGE_PERCENT,
        historical_hourly_conditions_field=historical_hourly_conditions.Field.CLOUD_COVERAGE_PERCENT,
        enhanced_forecast_field=enhanced_forecast.Field.CLOUD_COVER,
    ),
    Field.DEW_POINT_TEMPERATURE: _TWCFieldTriple(
        cleaned_observations_field=cleaned_observations.Field.SURFACE_DEWPOINT_TEMPERATURE_CELSIUS,
        historical_hourly_conditions_field=historical_hourly_conditions.Field.TEMPERATURE_DEW_POINT,
        enhanced_forecast_field=enhanced_forecast.Field.TEMPERATURE_DEW_POINT,
    ),
    Field.RELATIVE_HUMIDITY: _TWCFieldTriple(
        cleaned_observations_field=cleaned_observations.Field.RELATIVE_HUMIDITY_PERCENT,
        historical_hourly_conditions_field=historical_hourly_conditions.Field.RELATIVE_HUMIDITY,
        enhanced_forecast_field=enhanced_forecast.Field.RELATIVE_HUMIDITY,
    ),
    Field.TEMPERATURE: _TWCFieldTriple(
        cleaned_observations_field=cleaned_observations.Field.SURFACE_TEMPERATURE_CELSIUS,
        historical_hourly_conditions_field=historical_hourly_conditions.Field.TEMPERATURE,
        enhanced_forecast_field=enhanced_forecast.Field.TEMPERATURE,
    ),
    Field.WIND_CHILL_TEMPERATURE: _TWCFieldTriple(
        cleaned_observations_field=cleaned_observations.Field.WIND_CHILL_TEMPERATURE_CELSIUS,
        historical_hourly_conditions_field=historical_hourly_conditions.Field.TEMPERATURE_WIND_CHILL,
        enhanced_forecast_field=enhanced_forecast.Field.TEMPERATURE_WIND_CHILL,
    ),
    Field.WIND_DIRECTION: _TWCFieldTriple(
        cleaned_observations_field=cleaned_observations.Field.WIND_DIRECTION_DEGREES,
        historical_hourly_conditions_field=historical_hourly_conditions.Field.WIND_DIRECTION,
        enhanced_forecast_field=enhanced_forecast.Field.WIND_DIRECTION,
    ),
    Field.WIND_SPEED: _TWCFieldTriple(
        cleaned_observations_field=cleaned_observations.Field.WIND_SPEED_KPH,
        historical_hourly_conditions_field=historical_hourly_conditions.Field.WIND_SPEED,
        enhanced_forecast_field=enhanced_forecast.Field.WIND_SPEED,
    ),
}


@dataclass
class _SourceCache:
    cleaned_observations_source: Source
    historical_hourly_conditions_source: Source
    enhanced_forecast_source: Source


_SOURCE_CACHE: MutableMapping[Tuple[UUID, MetarStation], _SourceCache] = {}


class TheWeatherCompany(TimeSeriesRecipe):
    """A recipe for creating time series that stitch together multiple The Weather Company data sources.

    This recipe is defined for a given METAR station and field. Because the three The Weather Company APIs involved in
    this recipe interpolate differently between METAR stations, we do not allow specification of arbitrary lat/long,
    which could result in a time series of poor quality. Instead, we require a particular METAR station.
    """

    def __init__(
        self, metar_station: MetarStation, field: Field, create_intermediate_time_series: bool = False
    ) -> None:
        self._metar_station = metar_station
        self._field = field
        self._create_intermediate_time_series = create_intermediate_time_series

    def create(self, project: Project, title: Optional[str] = None, description: Optional[str] = None) -> TimeSeries:
        """Creates a new time series in the given project containing weather data for this recipe.

        This method will create the following resources in the given project:

        - A source specifying the `CleanedObservations` connector.
        - A source specifying the `HistoricalHourlyConditions` connector.
        - A source specifying the `EnhancedForecast` connector.
        - For each source, an optional intermediate time series with one layer from that source.
        - A time series with three layers:
          - One from the cleaned observations node, with no start offset and an end offset of -23 hours (exclusive)
          - One from the historical hourly conditions node, with a start offset of -23 hours (inclusive) and an
            end offset of 1 hour (exclusive)
          - One from the enhanced forecast node, with a start offset of 1 hour (inclusive) and no end offset

        When invoking this method multiple times for the same project, with recipes created for the same METAR station,
        the same sources will be reused.

        Note: when invoked with `create_intermediate_time_series=True`, the final time series's three layers link to the
        three intermediate time series nodes.

        Args:
            project: the project in which to create the time series and attendant sources
            title: the title of the time series; if none is given, will default to a combination of the field name and
                METAR station specified in the recipe
            description: the description of the time series

        Returns:
            the created time series
        """
        latitude = self._metar_station.latitude
        longitude = self._metar_station.longitude

        # Cache/look up sources by project + METAR station. This allows low-effort reuse of TWC sources within a single
        # client process.
        source_cache_key = (project.uuid, self._metar_station)

        # Create the three sources.
        if source_cache_key not in _SOURCE_CACHE:
            cleaned_observations_fields = [
                field_triple.cleaned_observations_field for field_triple in _FIELD_MAPPINGS.values()
            ]
            cleaned_observations_source = Source.create(
                project=project,
                title=f"CO ({self._metar_station.name})",
                connector=cleaned_observations.CleanedObservations(
                    latitude=latitude, longitude=longitude, fields=cleaned_observations_fields
                ),
            )

            historical_hourly_conditions_fields = [
                field_triple.historical_hourly_conditions_field for field_triple in _FIELD_MAPPINGS.values()
            ]
            historical_hourly_conditions_source = Source.create(
                project=project,
                title=f"HHC ({self._metar_station.name})",
                connector=historical_hourly_conditions.HistoricalHourlyConditions(
                    latitude=latitude, longitude=longitude, fields=historical_hourly_conditions_fields
                ),
            )

            enhanced_forecast_fields = [
                field_triple.enhanced_forecast_field for field_triple in _FIELD_MAPPINGS.values()
            ]
            enhanced_forecast_source = Source.create(
                project=project,
                title=f"EF ({self._metar_station.name})",
                connector=enhanced_forecast.EnhancedForecast(
                    latitude=latitude, longitude=longitude, fields=enhanced_forecast_fields
                ),
            )

            _SOURCE_CACHE[source_cache_key] = _SourceCache(
                cleaned_observations_source=cleaned_observations_source,
                historical_hourly_conditions_source=historical_hourly_conditions_source,
                enhanced_forecast_source=enhanced_forecast_source,
            )

        # Create the intermediate time series.
        if self._create_intermediate_time_series:
            co_time_series = TimeSeries.create(
                project=project,
                title=f"CO {self._field.value} ({self._metar_station.name})",
                sample_period=_THE_WEATHER_COMPANY_DATA_SAMPLE_PERIOD,
                cell_shape=(),
                coordinate_labels=(),
                axis_labels=(),
            )
            TimeSeriesLayer.create(
                project=project,
                time_series=co_time_series,
                node=_SOURCE_CACHE[source_cache_key].cleaned_observations_source,
                order=0,
                output_index=0,
                label_indexer=_FIELD_MAPPINGS[self._field].cleaned_observations_field,
            )

            hhc_time_series = TimeSeries.create(
                project=project,
                title=f"HHC {self._field.value} ({self._metar_station.name})",
                sample_period=_THE_WEATHER_COMPANY_DATA_SAMPLE_PERIOD,
                cell_shape=(),
                coordinate_labels=(),
                axis_labels=(),
            )
            TimeSeriesLayer.create(
                project=project,
                time_series=hhc_time_series,
                node=_SOURCE_CACHE[source_cache_key].historical_hourly_conditions_source,
                order=0,
                output_index=0,
                label_indexer=_FIELD_MAPPINGS[self._field].historical_hourly_conditions_field,
            )

            ef_time_series = TimeSeries.create(
                project=project,
                title=f"EF {self._field.value} ({self._metar_station.name})",
                sample_period=_THE_WEATHER_COMPANY_DATA_SAMPLE_PERIOD,
                cell_shape=(),
                coordinate_labels=(),
                axis_labels=(),
            )
            TimeSeriesLayer.create(
                project=project,
                time_series=ef_time_series,
                node=_SOURCE_CACHE[source_cache_key].enhanced_forecast_source,
                order=0,
                output_index=0,
                label_indexer=_FIELD_MAPPINGS[self._field].enhanced_forecast_field,
            )

            co_upstream_node: Union[Source, TimeSeries] = co_time_series
            hhc_upstream_node: Union[Source, TimeSeries] = hhc_time_series
            ef_upstream_node: Union[Source, TimeSeries] = ef_time_series

            co_label_indexer = None
            hhc_label_indexer = None
            ef_label_indexer = None
        else:
            co_upstream_node = _SOURCE_CACHE[source_cache_key].cleaned_observations_source
            hhc_upstream_node = _SOURCE_CACHE[source_cache_key].historical_hourly_conditions_source
            ef_upstream_node = _SOURCE_CACHE[source_cache_key].enhanced_forecast_source

            co_label_indexer = _FIELD_MAPPINGS[self._field].cleaned_observations_field
            hhc_label_indexer = _FIELD_MAPPINGS[self._field].historical_hourly_conditions_field
            ef_label_indexer = _FIELD_MAPPINGS[self._field].enhanced_forecast_field

        # Create the time series.
        default_title = f"{self._field.value} ({self._metar_station.name})"
        time_series = TimeSeries.create(
            project=project,
            title=title or default_title,
            sample_period=_THE_WEATHER_COMPANY_DATA_SAMPLE_PERIOD,
            cell_shape=(),
            coordinate_labels=(),
            axis_labels=(),
            description=description,
        )
        TimeSeriesLayer.create(
            project=project,
            time_series=time_series,
            node=co_upstream_node,
            order=0,
            output_index=0,
            label_indexer=co_label_indexer,
            end_timing=_HISTORICAL_APIS_CUTOVER_BOUNDARY,
        )
        TimeSeriesLayer.create(
            project=project,
            time_series=time_series,
            node=hhc_upstream_node,
            order=1,
            output_index=0,
            label_indexer=hhc_label_indexer,
            start_timing=_HISTORICAL_APIS_CUTOVER_BOUNDARY,
            end_timing=_HISTORICAL_TO_FORECAST_CUTOVER_BOUNDARY,
        )
        TimeSeriesLayer.create(
            project=project,
            time_series=time_series,
            node=ef_upstream_node,
            order=2,
            output_index=0,
            label_indexer=ef_label_indexer,
            start_timing=_HISTORICAL_TO_FORECAST_CUTOVER_BOUNDARY,
        )

        return time_series



from typing import Any, Dict, Generator

import pytest
from respx import MockRouter

import myst
from myst.auth.credentials import FakeCredentials


@pytest.fixture(scope="module", autouse=True)
def use_fake_credentials() -> None:
    myst.set_client(myst.Client(credentials=FakeCredentials()))


@pytest.fixture(autouse=True)
def mock_api() -> Generator[MockRouter, None, None]:
    with MockRouter(base_url=f"{myst.settings.MYST_API_HOST}/v1alpha2/") as mock_router:
        yield mock_router


@pytest.fixture
def example_project() -> Dict[str, Any]:
    return dict(
        object="Project",
        uuid="7c1ccecc-f03b-4f33-95bf-581ee96a900e",
        create_time="2021-09-15T23:35:00Z",
        update_time=None,
        organization="f02caef7-07a4-4c47-8901-9680c2124f2e",
        organization_sharing_enabled=False,
        owner="e4769980-e154-4b10-96b7-ca9f72671843",
        title="uncontradictious-crassina-talemonger-euphroe",
        description="Some test project.",
        creator="e4769980-e154-4b10-96b7-ca9f72671843",
        deploy_status="NEW",
    )


@pytest.fixture
def example_source(example_project: Dict[str, Any]) -> Dict[str, Any]:
    return dict(
        object="Node",
        uuid="b13ffa7b-1bf2-4ed5-bf96-8701df9e1b1c",
        create_time="2021-09-15T22:36:17Z",
        update_time=None,
        organization="99799474-bc23-4026-ae88-89471a077b69",
        owner="fb80ae8a-dc73-45f3-9a49-ca57202b8fdf",
        type="Source",
        title="Forecasted Weather in SF",
        description=None,
        project=example_project["uuid"],
        creator="fb80ae8a-dc73-45f3-9a49-ca57202b8fdf",
        deploy_status="NEW",
        connector_uuid="1aeaadec-6912-422a-a4fc-5d6765391c17",
        parameters=dict(latitude=37.7948, longitude=-122.3986, fields=["relativeHumidity", "temperature"]),
    )


@pytest.fixture
def example_model(example_project: Dict[str, Any]) -> Dict[str, Any]:
    return dict(
        object="Node",
        uuid="36cad5b7-6ea8-4e08-9546-d28218e2486f",
        create_time="2021-09-15T22:36:35Z",
        update_time=None,
        organization="99799474-bc23-4026-ae88-89471a077b69",
        owner="fb80ae8a-dc73-45f3-9a49-ca57202b8fdf",
        type="Model",
        title="An XGBoost Model",
        description=None,
        project=example_project["uuid"],
        creator="fb80ae8a-dc73-45f3-9a49-ca57202b8fdf",
        deploy_status="NEW",
        connector_uuid="b78ff94a-27b1-4986-a88a-536661239bb2",
        input_specs_schema=dict(foo="bar"),
        parameters=dict(
            num_boost_round=10,
            max_depth=6,
            min_child_weight=1,
            learning_rate=0.3,
            subsample=1.0,
            colsample_bytree=1.0,
            colsample_bylevel=1.0,
            colsample_bynode=1.0,
            gamma=0.0,
            reg_alpha=0.0,
            reg_lambda=1.0,
        ),
    )


@pytest.fixture
def example_operation(example_project: Dict[str, Any]) -> Dict[str, Any]:
    return dict(
        object="Node",
        uuid="d0f21c98-b6c5-42ba-98c8-6d336f6be4be",
        create_time="2021-09-15T22:36:30Z",
        update_time=None,
        organization="99799474-bc23-4026-ae88-89471a077b69",
        owner="fb80ae8a-dc73-45f3-9a49-ca57202b8fdf",
        type="Operation",
        title="3H Rolling Mean",
        description="Takes a rolling mean of the upstream series",
        project=example_project["uuid"],
        creator="fb80ae8a-dc73-45f3-9a49-ca57202b8fdf",
        deploy_status="NEW",
        connector_uuid="e4bebd6d-d9d2-4103-8c94-95fa162eddc0",
        input_specs_schema=dict(foo="bar"),
        parameters=dict(
            rolling_window_parameters=dict(
                window_period="PT3H", min_periods=1, centered=False, aggregation_function="mean"
            ),
            differencing_parameters=None,
            shift_parameters=None,
        ),
    )


@pytest.fixture
def example_time_series(example_project: Dict[str, Any]) -> Dict[str, Any]:
    return dict(
        object="Node",
        uuid="221c85ec-0bf1-477c-a820-ce0da7dd6f04",
        create_time="2021-09-15T22:36:42Z",
        update_time=None,
        organization="99799474-bc23-4026-ae88-89471a077b69",
        owner="fb80ae8a-dc73-45f3-9a49-ca57202b8fdf",
        type="TimeSeries",
        title="The thing we want to predict",
        description="This is a time series",
        project=example_project["uuid"],
        creator="fb80ae8a-dc73-45f3-9a49-ca57202b8fdf",
        deploy_status="NEW",
        sample_period="PT1H",
        cell_shape=[2, 3],
        coordinate_labels=[["x1", "x2"], ["y1", "y2", "y3"]],
        axis_labels=["x", "y"],
    )


@pytest.fixture
def example_model_input(example_model: Dict[str, Any]) -> Dict[str, Any]:
    return dict(
        object="Edge",
        uuid="ceb01ff3-badf-4829-9322-911168d502aa",
        create_time="2021-09-15T22:36:36Z",
        update_time=None,
        type="Input",
        project=example_model["project"],
        downstream_node=example_model["uuid"],
        upstream_node="e6af86ed-850c-484d-898e-72a53b4ffe7f",
        output_index=0,
        label_indexer=None,
        group_name="features",
    )


@pytest.fixture
def example_operation_input(example_operation: Dict[str, Any]) -> Dict[str, Any]:
    return dict(
        object="Edge",
        uuid="8046a461-b58d-40fe-ae4a-079ff2f9c213",
        create_time="2021-09-15T22:36:31Z",
        update_time=None,
        type="Input",
        project=example_operation["project"],
        downstream_node=example_operation["uuid"],
        upstream_node="d4d750fe-f727-4073-b6d9-8c643476ac93",
        output_index=0,
        label_indexer=None,
        group_name="operands",
    )


@pytest.fixture
def example_layer(example_time_series: Dict[str, Any]) -> Dict[str, Any]:
    return dict(
        object="Edge",
        uuid="e09ee995-1aaa-4ad3-b91a-925b8f0a4679",
        create_time="2021-09-15T22:36:44Z",
        update_time=None,
        type="Layer",
        project=example_time_series["project"],
        downstream_node=example_time_series["uuid"],
        upstream_node="36cad5b7-6ea8-4e08-9546-d28218e2486f",
        output_index=0,
        label_indexer=None,
        order=0,
        start_timing=None,
        end_timing=None,
    )


@pytest.fixture
def example_backtest(example_project: Dict[str, Any], example_model: Dict[str, Any]) -> Dict[str, Any]:
    fit_start_timing_str = "-P1Y"
    fit_end_timing_str = "PT0H"
    predict_start_timing_str = "PT1H"
    predict_end_timing_str = "PT6H"

    return dict(
        object="Backtest",
        uuid="6c219892-5d88-40d8-85d4-56170f752336",
        project=example_project["uuid"],
        create_time="2021-09-15T22:36:44Z",
        update_time=None,
        creator="fb80ae8a-dc73-45f3-9a49-ca57202b8fdf",
        owner="fb80ae8a-dc73-45f3-9a49-ca57202b8fdf",
        title="example_backtest",
        description=None,
        model=example_model["uuid"],
        test_start_time="2020-01-01T00:00:00Z",
        test_end_time="2021-01-01T00:00:00Z",
        fit_start_timing=dict(object="Timing", type="RelativeTiming", offset=fit_start_timing_str, time_zone=None),
        fit_end_timing=dict(object="Timing", type="RelativeTiming", offset=fit_end_timing_str, time_zone=None),
        fit_reference_timing=dict(object="Timing", type="CronTiming", cron_expression="0 0 * * 0", time_zone="UTC"),
        predict_start_timing=dict(
            object="Timing", type="RelativeTiming", offset=predict_start_timing_str, time_zone="UTC"
        ),
        predict_end_timing=dict(object="Timing", type="RelativeTiming", offset=predict_end_timing_str, time_zone="UTC"),
        predict_reference_timing=dict(object="Timing", type="CronTiming", cron_expression="0 * * * *", time_zone="UTC"),
    )


@pytest.fixture
def example_hpo(example_project: Dict[str, Any], example_model: Dict[str, Any]) -> Dict[str, Any]:
    fit_start_timing_str = "-P1Y"
    fit_end_timing_str = "PT0H"
    predict_start_timing_str = "PT1H"
    predict_end_timing_str = "PT6H"

    return dict(
        object="HPO",
        uuid="c72c0a3d-7690-4b22-bafb-267437a1e648",
        project=example_project["uuid"],
        create_time="2021-09-15T22:36:44Z",
        update_time=None,
        creator="fb80ae8a-dc73-45f3-9a49-ca57202b8fdf",
        owner="fb80ae8a-dc73-45f3-9a49-ca57202b8fdf",
        title="example_hpo",
        description=None,
        model=example_model["uuid"],
        test_start_time="2020-01-01T00:00:00Z",
        test_end_time="2021-01-01T00:00:00Z",
        fit_start_timing=dict(object="Timing", type="RelativeTiming", offset=fit_start_timing_str, time_zone=None),
        fit_end_timing=dict(object="Timing", type="RelativeTiming", offset=fit_end_timing_str, time_zone=None),
        fit_reference_timing=dict(object="Timing", type="CronTiming", cron_expression="0 0 * * 0", time_zone="UTC"),
        predict_start_timing=dict(
            object="Timing", type="RelativeTiming", offset=predict_start_timing_str, time_zone="UTC"
        ),
        predict_end_timing=dict(object="Timing", type="RelativeTiming", offset=predict_end_timing_str, time_zone="UTC"),
        predict_reference_timing=dict(object="Timing", type="CronTiming", cron_expression="0 * * * *", time_zone="UTC"),
        search_algorithm=dict(
            object="SearchAlgorithm",
            type="Hyperopt",
            metric="rmse",
            num_trials=12,
            max_concurrent_trials=3,
            algorithm="TPE",
            n_startup_jobs=1,
        ),
        search_space=dict(max_depth=dict(object="Sampler", type="Uniform", lower=1, upper=12)),
    )


@pytest.fixture
def example_result() -> Dict[str, Any]:
    return dict(
        object="BacktestResult",
        uuid="37faf8b5-5a4e-40e5-b90a-439a9743f921",
        create_time="2021-03-17T00:00:09Z",
        start_time="2021-03-17T00:00:10Z",
        end_time="2021-03-17T00:03:42Z",
        result_url="https://storage.cloud.google.com/backtest-result/37faf8b5-5a4e-40e5-b90a-439a9743f921.json",
        metrics=dict(mape=1.4, mae=43, rmse=1.39348),
    )


@pytest.fixture
def example_deployment(example_project: Dict[str, Any]) -> Dict[str, Any]:
    return dict(
        object="Deployment",
        project=example_project["uuid"],
        uuid="3fa85f64-5717-4562-b3fc-2c963f66afa6",
        create_time="2021-01-01T01:23:45Z",
        title="My deployment",
        activate_time="2021-01-01T01:23:45Z",
        creator=example_project["creator"],
    )


@pytest.fixture
def example_hpo_result() -> Dict[str, Any]:
    return dict(
        data=[
            dict(
                object="HPOResult",
                uuid="37faf8b5-5a4e-40e5-b90a-439a9743f921",
                create_time="2021-03-17T00:00:09Z",
                metric="rmse",
                best_trial=dict(
                    parameters=dict(p1=3, p2=4),
                    metrics=dict(mae=1.4, rmse=43, mape=0.2),
                    create_time="2021-03-17T00:00:09Z",
                    backtest_result_url="mock-url",
                ),
                trials=[
                    dict(
                        parameters=dict(p1=3, p2=4),
                        metrics=dict(mae=1.4, rmse=43, mape=0.2),
                        create_time="2021-03-17T00:00:09Z",
                        backtest_result_url="mock-url",
                    )
                ],
            )
        ],
        has_more=False,
    )


@pytest.fixture
def example_hpo_job(example_hpo: Dict[str, Any]) -> Dict[str, Any]:
    return dict(
        object="HPOJob",
        uuid="c72c0a3d-7690-4b22-bafb-267437a1e647",
        hpo=example_hpo["uuid"],
        state="SUCCESS",
        create_time="2021-03-17T00:00:09Z",
        schedule_time="2021-03-17T00:00:09Z",
        num_trials_completed=0,
        creator="626e0364-c334-4148-958b-aeac1ad2d6d6",
        detail="",
    )


@pytest.fixture
def example_model_fit_job(example_project: Dict[str, Any], example_model: Dict[str, Any]) -> Dict[str, Any]:
    return dict(
        object="NodeJob",
        type="ModelFitJob",
        uuid="c72c0a3d-7690-4b22-bafb-267437a1e649",
        project=example_project["uuid"],
        node=example_model["uuid"],
        state="SUCCESS",
        create_time="2021-03-17T00:00:09Z",
        start_time="2021-03-17T00:00:10Z",
        end_time="2021-03-17T00:03:42Z",
        as_of_time="2021-03-17T00:00:00Z",
        result=None,
        errors=[],
    )


@pytest.fixture
def example_time_series_run_job(example_project: Dict[str, Any], example_time_series: Dict[str, Any]) -> Dict[str, Any]:
    return dict(
        object="NodeJob",
        type="NodeRunJob",
        uuid="c72c0a3d-7690-4b22-bafb-267437a1e64a",
        project=example_project["uuid"],
        node=example_time_series["uuid"],
        state="SUCCESS",
        create_time="2021-03-17T00:00:09Z",
        start_time="2021-03-17T00:00:10Z",
        end_time="2021-03-17T00:03:42Z",
        as_of_time="2021-03-17T00:00:00Z",
        result=None,
        errors=[],
    )



from typing import Optional
from uuid import UUID

from myst.models.types import ItemOrSlice
from myst.resources.resource import Resource


class Edge(Resource):
    """An edge between two nodes in a project graph.

    Attributes:
        project: identifier of the project this edge belongs to
        upstream_node: the identifier of the node data flows into this edge from
        downstream_node: the identifier of the node data flows out of this edge to
        output_index: which time dataset, out of the sequence of upstream time datasets, to pass to this edge
        label_indexer: the slice of the upstream data to pass to this edge
    """

    project: UUID
    upstream_node: UUID
    downstream_node: UUID
    output_index: int
    label_indexer: Optional[ItemOrSlice] = None



from myst.client import Client
from myst.openapi.models.model_create import ModelCreate
from myst.openapi.models.model_get import ModelGet


def request_sync(client: Client, project_uuid: str, json_body: ModelCreate) -> ModelGet:
    """Creates a model."""

    return client.request(
        method="post", path=f"/projects/{project_uuid}/models/", response_class=ModelGet, request_model=json_body
    )



from myst.models import base_model
from myst.openapi.models.operation import Operation


class PermissionGet(base_model.BaseModel):
    """Schema for permission get responses."""

    operation: Operation



from myst.client import Client
from myst.openapi.models.node_job_create import NodeJobCreate
from myst.openapi.models.node_run_job_get import NodeRunJobGet


def request_sync(client: Client, project_uuid: str, uuid: str, json_body: NodeJobCreate) -> NodeRunJobGet:
    """Create an ad hoc run job for a time series."""

    return client.request(
        method="post",
        path=f"/projects/{project_uuid}/time_series/{uuid}:run",
        response_class=NodeRunJobGet,
        request_model=json_body,
    )



import abc
import pathlib
from abc import abstractmethod
from typing import Optional

from google.auth.transport.requests import Request as GoogleRequest
from google.oauth2.credentials import Credentials as GoogleCredentials
from google.oauth2.service_account import IDTokenCredentials
from google_auth_oauthlib.flow import InstalledAppFlow

import myst
from myst.auth.user_credentials_cache import UserCredentialsCache
from myst.settings import PROJECT_ROOT, settings


class Credentials(abc.ABC):
    """Abstraction around various types of credentials which enable use of the Myst API."""

    @property
    @abstractmethod
    def token(self) -> str:
        """Validated Bearer token using these credentials."""
        raise NotImplementedError()


class GoogleServiceAccountCredentials(Credentials):
    """Credentials that use a service account.

    API docs are here:
    https://google-auth.readthedocs.io/en/master/reference/google.oauth2.service_account.html
    """

    def __init__(self, key_file_path: Optional[pathlib.Path] = None, target_audience: Optional[str] = None):
        super().__init__()

        self._credentials = IDTokenCredentials.from_service_account_file(
            filename=key_file_path or settings.MYST_APPLICATION_CREDENTIALS,
            target_audience=target_audience or settings.MYST_API_HOST,
        )

    @property
    def credentials(self) -> IDTokenCredentials:
        """Underlying Google ID token credentials."""
        return self._credentials

    @property
    def token(self) -> str:
        """Validated Bearer token using these credentials."""
        if not self.credentials.valid:
            self.credentials.refresh(GoogleRequest())

        return self.credentials.token


class GoogleConsoleCredentials(Credentials):
    """Credentials created by sending the user through a local OAuth flow.

    API docs are here:
    https://google-auth-oauthlib.readthedocs.io/en/latest/reference/google_auth_oauthlib.flow.html
    """

    CLIENT_SECRET_FILE = PROJECT_ROOT / "myst" / "data" / "google_oauth_client_not_so_secret.json"

    def __init__(self, use_console: bool = False, use_cache: bool = True) -> None:
        super().__init__()

        self._credentials = self._get_credentials(use_console=use_console, use_cache=use_cache)

    @property
    def credentials(self) -> GoogleCredentials:
        """Underlying Google credentials."""
        return self._credentials

    @property
    def token(self) -> str:
        """Validated Bearer token using these credentials."""
        if not self.credentials.valid:
            self.credentials.refresh(GoogleRequest())

        return self.credentials.id_token

    def _get_credentials(self, use_console: bool, use_cache: bool) -> GoogleCredentials:
        """Obtains credentials via cache or OAuth flow."""
        user_credentials_cache = UserCredentialsCache()

        # Note that this will also return `None` if user credentials were not found or couldn't be loaded from cache.
        user_credentials = user_credentials_cache.load() if use_cache else None

        if user_credentials:
            return user_credentials

        # Couldn't find user credentials in cache or explicitly told not to use cache, so perform the OAuth2.0 dance.
        flow = InstalledAppFlow.from_client_secrets_file(
            client_secrets_file=self.CLIENT_SECRET_FILE,
            scopes=["openid", "https://www.googleapis.com/auth/userinfo.email"],
        )
        user_credentials = flow.run_console() if use_console else flow.run_local_server()

        # Write newly fetched user credentials to cache.
        if use_cache:
            user_credentials_cache.save(user_credentials=user_credentials)

        return user_credentials


class FakeCredentials(Credentials):
    """Fake credentials for use in tests."""

    @property
    def token(self) -> str:
        return "fake-token"


def authenticate(credentials: Optional[Credentials] = None, use_console: bool = False, use_cache: bool = True) -> None:
    """Authenticates the global client with user credentials.

    Args:
        credentials: custom credentials to use instead of getting credentials from user account; if provided, this
            method will prefer these over other authentication methods
        use_console: whether or not to use the console for the user account authorization flow
        use_cache: whether or not to reuse cached user credentials
    """
    if credentials is None:
        credentials = GoogleConsoleCredentials(use_console=use_console, use_cache=use_cache)

    myst.get_client().authenticate(credentials=credentials)


def authenticate_with_service_account(key_file_path: Optional[pathlib.Path] = None) -> None:
    """Authenticates the global client with service account credentials.

    Args:
        key_file_path: absolute path to service account key file; if this is not given, this will look
            for the key file path in the `MYST_APPLICATION_CREDENTIALS` environment variable
    """
    client = myst.get_client()
    creds = GoogleServiceAccountCredentials(key_file_path=key_file_path, target_audience=client.base_url)
    client.authenticate(credentials=creds)



from typing import Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.hpo_metric import HPOMetric


class HyperoptCreate(base_model.BaseModel):
    """Hyperopt search algorithm schema for create requests.

    See https://hyperopt.github.io/hyperopt/ for `hyperopt` reference."""

    object_: Optional[Literal["SearchAlgorithm"]] = Field(..., alias="object")
    type: Optional[Literal["Hyperopt"]] = "Hyperopt"
    metric: Optional[HPOMetric] = HPOMetric.RMSE
    num_trials: Optional[int] = 10
    max_concurrent_trials: Optional[int] = 1
    algorithm: Optional[Literal["TPE"]] = "TPE"
    n_startup_jobs: Optional[int] = None



from functools import singledispatch
from typing import Any, List, Mapping, Optional, Tuple, TypeVar, Union
from uuid import UUID


class Unset:
    pass


UNSET = Unset()
UUIDOrStr = Union[str, UUID]
IntOrStr = Union[int, str]
Shape = Tuple[int, ...]
CoordinateLabels = Tuple[Tuple[IntOrStr, ...], ...]
AxisLabels = Tuple[IntOrStr, ...]
Metadata = Mapping[str, Union[str, int, float, bool, None]]

# Note: This should be a recursive type alias but `mypy` doesn't support that, so use `List[Any]` instead to allow for
#   an arbitrary amount of nesting.
ItemOrSlice = Union[Tuple[Union[IntOrStr, List[Any]], ...], List[Any], IntOrStr]

T = TypeVar("T")
OptionalArgument = Union[Optional[T], Unset]


@singledispatch
def to_uuid(uuid: UUIDOrStr) -> UUID:
    raise NotImplementedError()


@to_uuid.register
def _uuid_to_uuid(uuid: UUID) -> UUID:
    return uuid


@to_uuid.register
def _str_to_uuid(uuid: str) -> UUID:
    return UUID(uuid)



from typing import List

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.hpo_metric import HPOMetric
from myst.openapi.models.hpo_trial import HPOTrial


class HPOResultGet(base_model.BaseModel):
    """Schema for model fit result get responses."""

    object_: Literal["HPOResult"] = Field(..., alias="object")
    uuid: str
    create_time: str
    metric: HPOMetric
    best_trial: HPOTrial
    trials: List[HPOTrial]



from myst.client import Client
from myst.openapi.models.hpo_get import HPOGet


def request_sync(client: Client, project_uuid: str, uuid: str) -> HPOGet:
    """Gets an HPO in the project."""

    return client.request(method="get", path=f"/projects/{project_uuid}/hpos/{uuid}", response_class=HPOGet)



from myst.client import Client
from myst.openapi.models.organization_get import OrganizationGet
from myst.openapi.models.organization_update import OrganizationUpdate


def request_sync(client: Client, uuid: str, json_body: OrganizationUpdate) -> OrganizationGet:
    """Updates an organization."""

    return client.request(
        method="patch", path=f"/organizations/{uuid}", response_class=OrganizationGet, request_model=json_body
    )



import enum
from typing import List
from uuid import UUID

from myst.connectors.source_connector import SourceConnector
from myst.core.time.time_delta import TimeDelta


@enum.unique
class Field(str, enum.Enum):
    AZIMUTH = "azimuth"
    ELEVATION = "elevation"
    ZENITH = "zenith"


class SolarPosition(SourceConnector):
    def __init__(self, sample_period: TimeDelta, latitude: float, longitude: float, fields: List[Field]) -> None:
        super().__init__(
            uuid=UUID("9bf1b9dd-fd14-4b1e-bd47-a238dbcfc551"),
            parameters=dict(sample_period=sample_period, latitude=latitude, longitude=longitude, fields=fields),
        )



from enum import Enum


class Operation(str, Enum):
    VIEW = "VIEW"
    EDIT = "EDIT"
    RESHARE = "RESHARE"
    TRANSFER_OWNERSHIP = "TRANSFER_OWNERSHIP"
    DELETE = "DELETE"

    def __str__(self) -> str:
        return str(self.value)



from typing import Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.hpo_job_state import HPOJobState


class HPOJobGet(base_model.BaseModel):
    """HPO job schema for get responses."""

    object_: Literal["HPOJob"] = Field(..., alias="object")
    uuid: str
    create_time: str
    hpo: str
    creator: str
    schedule_time: str
    state: HPOJobState
    num_trials_completed: int
    update_time: Optional[str] = None
    detail: Optional[str] = None
    result: Optional[str] = None



from enum import Enum


class JobState(str, Enum):
    PENDING = "PENDING"
    STARTED = "STARTED"
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"
    UPSTREAM_FAILURE = "UPSTREAM_FAILURE"

    def __str__(self) -> str:
        return str(self.value)



from myst.client import Client
from myst.openapi.models.deployment_create import DeploymentCreate
from myst.openapi.models.deployment_get import DeploymentGet


def request_sync(client: Client, project_uuid: str, json_body: DeploymentCreate) -> DeploymentGet:
    """Creates a deployment for a given project, deactivating any currently active deployment."""

    return client.request(
        method="post",
        path=f"/projects/{project_uuid}/deployments/",
        response_class=DeploymentGet,
        request_model=json_body,
    )



from typing import Any, List, Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.deploy_status import DeployStatus


class TimeSeriesUpdate(base_model.BaseModel):
    """Schema for time series update requests."""

    object_: Optional[Literal["Node"]] = Field(..., alias="object")
    uuid: Optional[str] = None
    create_time: Optional[str] = None
    update_time: Optional[str] = None
    organization: Optional[str] = None
    owner: Optional[str] = None
    type: Optional[Literal["TimeSeries"]] = None
    title: Optional[str] = None
    description: Optional[str] = None
    project: Optional[str] = None
    deploy_status: Optional[DeployStatus] = None
    sample_period: Optional[str] = None
    cell_shape: Optional[List[Any]] = None
    coordinate_labels: Optional[List[Any]] = None
    axis_labels: Optional[List[Any]] = None



from typing import TYPE_CHECKING, List, Union
from uuid import UUID

from myst.adapters.utils import get_resource_uuid
from myst.client import get_client
from myst.models.types import UUIDOrStr, to_uuid
from myst.openapi.api.projects.hpos.jobs import get_hpo_job, list_hpo_jobs
from myst.openapi.models.hpo_job_state import HPOJobState
from myst.resources.resource import Resource

if TYPE_CHECKING:  # Avoid circular imports.
    from myst.resources.hpo import HPO
    from myst.resources.project import Project


class HPOJob(Resource):

    hpo: UUID
    state: HPOJobState
    num_trials_completed: int

    @classmethod
    def get(cls, project: Union["Project", UUIDOrStr], hpo: Union["HPO", UUIDOrStr], uuid: UUIDOrStr) -> "HPOJob":
        """Gets a specific HPO job by its identifier."""
        hpo_job = get_hpo_job.request_sync(
            client=get_client(),
            project_uuid=str(get_resource_uuid(project)),
            hpo_uuid=str(get_resource_uuid(hpo)),
            uuid=str(to_uuid(uuid)),
        )

        return HPOJob.parse_obj(hpo_job.dict())

    @classmethod
    def list(cls, project: Union["Project", UUIDOrStr], hpo: Union["HPO", UUIDOrStr]) -> List["HPOJob"]:
        """Lists all jobs for the HPO."""
        hpo_jobs = list_hpo_jobs.request_sync(client=get_client(), project_uuid=str(get_resource_uuid(project)))

        return [
            HPOJob.parse_obj(hpo_job) for hpo_job in hpo_jobs.data if str(hpo_job.hpo) == str(get_resource_uuid(hpo))
        ]

    def is_completed(self) -> bool:
        """Returns `True` if the job has completed.

        Completed jobs include both successful and failed jobs.

        Returns:
            whether the job has completed
        """
        return self.state in [HPOJobState.SUCCESS, HPOJobState.FAILURE]



from myst.client import Client
from myst.openapi.models.alias_get import AliasGet


def request_sync(client: Client, project_uuid: str) -> AliasGet:
    """Create an alias for the project."""

    return client.request(method="post", path=f"/alias/-/project/{project_uuid}", response_class=AliasGet)



from typing import List, Union

from myst.models import base_model
from myst.openapi.models.model_fit_job_get import ModelFitJobGet
from myst.openapi.models.node_run_job_get import NodeRunJobGet


class ProjectNodeJobList(base_model.BaseModel):
    """Project job list schema."""

    data: List[Union[NodeRunJobGet, ModelFitJobGet]]
    has_more: bool



from myst.client import Client
from myst.openapi.models.permission_list import PermissionList
from myst.openapi.models.shareable_resource_share import ShareableResourceShare


def request_sync(client: Client, uuid: str, json_body: ShareableResourceShare) -> PermissionList:
    """Shares a project."""

    return client.request(
        method="post", path=f"/projects/{uuid}:share", response_class=PermissionList, request_model=json_body
    )



from typing import List

from myst.models import base_model
from myst.openapi.models.backtest_get import BacktestGet


class BacktestList(base_model.BaseModel):
    """Schema for backtest list responses."""

    data: List[BacktestGet]
    has_more: bool
